from lpsim.agents import InteractionAgent
from lpsim import Deck, Match, MatchState
from tests.utils_for_test import (
    check_hp,
    check_usage,
    get_pidx_cidx,
    get_random_state,
    get_test_id_from_command,
    make_respond,
    set_16_omni,
)


def test_joy_4_2_and_cost_change_cards():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 0 0 1",
            "skill 0 1 14 13",
            "skill 0 1 11 10",
            "end",
            "end",
            "end",
            "end",
        ],
        [
            "sw_card",
            "choose 1",
            "sw_char 2 15",
            "end",
            "end",
            "end",
            "end",
            "TEST 1 p0c0 app",
            "TEST 1 p0c1 app",
            "TEST 1 p0c2 app",
            "card 0 0 1",
            "TEST 1 p1c0 app",
            "TEST 1 p1c1 app",
            "TEST 1 p1c2 app",
            "TEST 2 p0 hand cost 3 3 4 4 2 4 4 3 3 4",
            "TEST 2 p1 hand cost 3 3 3 3 4 4 3 4 4",
            "TEST 3 p0 hand 5 charge 2",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.2
        character:Nahida
        character:Xiangling
        character:Mona
        Joyous Celebration
        Mirror Cage
        Mirror Cage@4.1
        Glorious Season
        Glorious Season@4.1
        Absorbing Prism
        Absorbing Prism@4.1
        Crossfire
        Crossfire@4.1
        The Overflow
        The Overflow@4.1
        Lands of Dandelion
        Lands of Dandelion@4.1
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                pidx, cidx = get_pidx_cidx(cmd)
                assert (
                    match.player_tables[pidx].characters[cidx].element_application == []
                )
            elif test_id == 2:
                pidx = int(cmd[2][1])
                hands = match.player_tables[pidx].hands
                costs = cmd[5:]
                assert len(hands) == len(costs)
                for h, c in zip(hands, costs):
                    c = int(c)
                    assert h.cost.total_dice_cost == c
            elif test_id == 3:
                assert match.player_tables[0].hands[5].cost.charge == 2
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_chongyun_yoimiya_beidou():
    cmd_records = [
        [
            "sw_card 0 2",
            "choose 0",
            "card 1 0 15 14 13 12",
            "TEST 2 p0 status 3",
            "TEST 2 p1 status 2",
            "skill 0 11 10 9",
            "TEST 1 4 10 10 4 10 10",
            "sw_char 1 8",
            "card 1 0 7 6",
            "TEST 3 p0c1 usage 3",
            "TEST 3 p1c1 usage 2",
            "skill 0 5 4 3",
            "TEST 1 4 6 10 4 6 10",
            "sw_char 2 2",
            "end",
            "end",
            "card 5 0 15 14 13",
            "sw_char 1 12",
            "skill 0 11 10",
            "skill 0 9 8",
            "skill 0 7 6 5",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "card 2 0 15 14 13",
            "skill 0 12 11 10",
            "sw_char 1 9",
            "card 3 0 8 7",
            "skill 0 6 5 4",
            "sw_char 2 3",
            "end",
            "end",
            "sw_char 1 1",
            "card 3 0 11 10 9",
            "skill 0 11 10 9",
            "skill 1 8 7 6",
            "skill 0 5 4",
            "TEST 1 4 1 3 4 1 3",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.2
        character:Chongyun
        character:Yoimiya
        character:Beidou
        Steady Breathing*2
        Steady Breathing@4.1*2
        Lightning Storm*2
        Lightning Storm@4.1*2
        Naganohara Meteor Swarm*2
        Naganohara Meteor Swarm@4.1*2
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                pidx = int(cmd[2][1])
                check_usage(match.player_tables[pidx].team_status, cmd[4:])
            elif test_id == 3:
                pidx, cidx = get_pidx_cidx(cmd)
                check_usage(match.player_tables[pidx].characters[cidx].status, cmd[4:])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_razor_sara_cyno():
    cmd_records = [
        [
            "sw_card 1 2",
            "choose 0",
            "card 0 0 15 14 13",
            "skill 1 12 11 10",
            "TEST 2 p0c1 charge 0",
            "TEST 2 p1c1 charge 1",
            "sw_char 1 9",
            "card 1 0 8 7 6",
            "sw_char 2 5",
            "card 0 0 4 3 2",
            "end",
            "skill 1 15 14 13",
            "sw_char 0 12",
            "end",
            "sw_char 1 15",
            "end",
            "TEST 1 2 2 2 4 5 4",
            "sw_char 2 15",
            "skill 0 14 13 12",
            "skill 1 11 10 9",
            "skill 1 8 7 6",
            "TEST 1 2 2 2 4 2 0",
            "end",
        ],
        [
            "sw_card 3 2",
            "choose 0",
            "card 1 0 15 14 13 12",
            "skill 1 11 10 9",
            "sw_char 1 8",
            "skill 0 7 6 5",
            "card 0 0 4 3 2 1",
            "TEST 1 4 8 9 4 5 10",
            "sw_char 2 0",
            "end",
            "card 1 0 15 14 13",
            "TEST 1 4 8 2 4 5 4",
            "end",
            "end",
            "skill 1 15 14 13",
            "end",
            "choose 1",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.2
        character:Razor
        character:Kujou Sara
        character:Cyno
        Awakening
        Awakening@4.1
        Sin of Pride
        Sin of Pride@4.1
        Featherfall Judgment
        Featherfall Judgment@4.1
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                pidx, cidx = get_pidx_cidx(cmd)
                assert match.player_tables[pidx].characters[cidx].charge == int(cmd[4])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_jean_kokomi_amber():
    cmd_records = [
        [
            "sw_card",
            "choose 2",
            "skill 0 15 14 13",
            "skill 0 12 11 10",
            "skill 2 9 8 7 6",
            "end",
            "TEST 1 10 10 5 10 10 10 10 8",
            "sw_char 1 15",
            "card 2 0 14 13 12",
            "sw_char 2 11",
            "skill 0 10 9 8",
            "sw_char 1 7",
            "skill 0 6 5 4",
            "TEST 1 10 5 5 10 10 4 10 8",
            "sw_char 0 3",
            "skill 0 2 1 0",
            "end",
            "skill 1 15 14 13",
            "card 1 0 12 11 10",
            "TEST 2 p1 summon 2",
            "end",
            "TEST 1 6 3 5 10 7 2 10 8",
            "sw_char 3 15",
            "end",
            "end",
            "sw_char 0 15",
            "skill 0 14 13 12",
            "skill 0 11 10 9",
            "skill 2 8 7 6",
            "TEST 2 p1 summon",
            "TEST 2 p0 summon 1",
            "end",
        ],
        [
            "sw_card",
            "choose 3",
            "skill 0 15 14 13",
            "skill 0 12 11 10",
            "skill 0 9 8 7",
            "skill 2 6 5 4 3",
            "end",
            "sw_char 1 15",
            "card 2 0 14 13 12",
            "sw_char 3 11",
            "skill 0 10 9 8",
            "sw_char 1 7",
            "skill 0 6 5 4",
            "sw_char 0 3",
            "skill 0 2 1 0",
            "end",
            "skill 1 15 14 13",
            "TEST 2 p0 summon 3",
            "TEST 2 p1 summon 2",
            "card 2 0 12 11 10",
            "end",
            "sw_char 2 15",
            "end",
            "end",
            "sw_char 0 15",
            "skill 0 14 13 12",
            "skill 0 11 10 9",
            "skill 2 8 7 6",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.2
        character:Sangonomiya Kokomi
        character:Amber
        character:Jean
        character:Jean@3.3
        Tamakushi Casket
        Tamakushi Casket@4.1
        Bunny Triggered
        Bunny Triggered@4.1
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:4], hps[4:]]
                check_hp(match, hps)
            elif test_id == 2:
                pidx = int(cmd[2][1])
                check_usage(match.player_tables[pidx].summons, cmd[4:])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_yanfei_4_2():
    cmd_records = [
        [
            "sw_card 0 1 3",
            "choose 0",
            "skill 1 15 14 13",
            "skill 1 12 11 10",
            "TEST 1 p0c0 status 2",
            "TEST 1 p1c1 status 1",
            "sw_char 1 9",
            "card 1 0 8 7 6",
            "end",
            "sw_char 0 15",
            "card 5 0 14 13 12",
            "TEST 2 p0 hand 5",
            "skill 0 11 10 9",
            "TEST 2 p0 hand 5",
            "skill 0 8 7 6",
            "card 4 0 5 4 3",
            "TEST 2 p0 hand 5",
            "end",
        ],
        [
            "sw_card 1 2 3 4",
            "choose 1",
            "skill 1 15 14 13",
            "skill 1 12 11 10",
            "sw_char 0 9",
            "sw_char 1 8",
            "card 3 0 7 6 5",
            "TEST 2 p1 hand 5",
            "sw_char 0 4",
            "end",
            "end",
            "choose 1",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.2
        character:Yanfei
        character:Yanfei@4.1
        Right of Final Interpretation*5
        Right of Final Interpretation@4.1*5
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                pidx, cidx = get_pidx_cidx(cmd)
                check_usage(match.player_tables[pidx].characters[cidx].status, cmd[4:])
            elif test_id == 2:
                pidx = int(cmd[2][1])
                assert len(match.player_tables[pidx].hands) == int(cmd[4])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_rhodeia_shenhe_itto():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 1 15 14 13",
            "TEST 2 p0 usage 2",
            "TEST 2 p1 usage 3",
            "sw_char 1 12",
            "skill 0 11 10 9",
            "skill 0 8 7 6",
            "skill 0 5 4 3",
            "skill 3 2 1 0",
            "TEST 1 8 5 10 8 3 10",
            "end",
            "skill 1 15 14 13",
            "skill 0 12 11 10",
            "skill 0 9 8 7",
            "sw_char 2 6",
            "sw_char 0 5",
            "end",
            "sw_char 1 15",
            "skill 3 14 13 12",
            "sw_char 2 11",
            "skill 1 10 9 8",
            "skill 1 7 6 5",
            "skill 1 4 3 2",
            "end",
            "TEST 1 1 3 1 3 0 10",
            "card 3 2 15",
            "card 2 1 14",
            "card 1 0 13",
            "skill 2 12 11 10",
            "end",
            "skill 0 15 14",
            "card 3 2 13",
            "card 3 1 12",
            "card 2 0 11",
            "sw_char 1 10",
            "TEST 1 5 1 5 3 0 10",
            "skill 0 9 8 7",
            "skill 1 6 5 4",
            "skill 1 3 2 1",
            "end",
            "skill 3 15 14 13",
            "TEST 1 5 1 5 3 0 3",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "skill 1 15 14 13",
            "sw_char 1 12",
            "skill 0 11 10 9",
            "skill 0 8 7 6",
            "skill 0 5 4 3",
            "skill 3 2 1 0",
            "end",
            "skill 1 15 14 13",
            "skill 0 12 11 10",
            "skill 0 9 8 7",
            "sw_char 0 6",
            "sw_char 1 5",
            "skill 3 4 3 2",
            "end",
            "choose 0",
            "skill 0 15 14 13",
            "TEST 1 1 3 10 3 0 10",
            "sw_char 2 12",
            "skill 1 11 10 9",
            "skill 1 8 7 6",
            "skill 1 5 4 3",
            "end",
            "skill 2 15 14 13",
            "TEST 1 3 5 3 3 0 6",
            "card 0 0 12",
            "card 3 0 11",
            "end",
            "sw_char 0 15",
            "card 5 1 14",
            "card 0 0 13",
            "TEST 1 3 5 3 3 0 10",
            "sw_char 2 12",
            "skill 0 11 10",
            "end",
            "card 9 1 15",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck1 = Deck.from_str(
        """
        default_version:4.2
        character:Shenhe
        character:Rhodeia of Loch
        character:Arataki Itto
        Mondstadt Hash Brown*30
        """
    )
    deck2 = Deck.from_str(
        """
        default_version:4.1
        character:Shenhe
        character:Rhodeia of Loch
        character:Arataki Itto
        Mondstadt Hash Brown*30
        """
    )
    match.set_deck([deck1, deck2])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                pidx = int(cmd[2][1])
                check_usage(match.player_tables[pidx].team_status, cmd[4:])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_kokomi_2():
    cmd_records = [
        [
            "sw_card",
            "choose 1",
            "skill 1 15 14 13",
            "sw_char 2 12",
            "skill 1 11 10 9",
            "sw_char 0 8",
            "skill 0 7 6 5",
            "skill 0 4 3 2",
            "end",
            "card 5 0 15 14 13",
            "TEST 1 p0 summon 1 1",
            "end",
        ],
        [
            "sw_card",
            "choose 1",
            "skill 1 15 14 13",
            "sw_char 0 12",
            "skill 1 11 10 9",
            "skill 1 8 7 6",
            "skill 2 5 4 3",
            "end",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.2
        character:Sangonomiya Kokomi
        character:Fischl
        character:Arataki Itto
        Tamakushi Casket*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                pidx = int(cmd[2][1])
                check_usage(match.player_tables[pidx].summons, cmd[4:])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    # import logging
    # logging.basicConfig(level = logging.INFO)
    test_joy_4_2_and_cost_change_cards()
    # test_chongyun_yoimiya_beidou()
    # test_razor_sara_cyno()
    # test_jean_kokomi_amber()
    # test_yanfei_4_2()
    # test_rhodeia_shenhe_itto()
    test_kokomi_2()



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import AttackerSummonBase

from ...modifiable_values import CostValue, DamageIncreaseValue
from ...event import RoundPrepareEventArguments

from ...action import Actions
from ...struct import Cost

from ...consts import (
    CostLabels,
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class SolarIsotoma_4_0(AttackerSummonBase):
    name: Literal["Solar Isotoma"] = "Solar Isotoma"
    version: Literal["4.0"] = "4.0"
    usage: int = 3
    max_usage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.GEO
    damage: int = 1

    attack_cost_decrease_usage: int = 1

    def renew(self, match: Any) -> None:
        super().renew(match)
        self.attack_cost_decrease_usage = 1

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        When in round prepare, reset usage
        """
        self.attack_cost_decrease_usage = 1
        return []

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        if self use plunging attack and has usage, decrease cost
        """
        if (
            self.attack_cost_decrease_usage > 0
            and value.cost.label & CostLabels.PLUNGING_ATTACK > 0
            and value.position.player_idx == self.position.player_idx
        ):
            # is self and has usage and plunging attack
            if value.cost.decrease_cost(None):
                if mode == "REAL":
                    self.attack_cost_decrease_usage -= 1
        return value

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["REAL", "TEST"]
    ) -> DamageIncreaseValue:
        """
        If has albedo equip talent, increase plunging attack by 1
        """
        if value.position.player_idx != self.position.player_idx:
            # not self
            return value
        found_talent_albedo = False
        characters = match.player_tables[self.position.player_idx].characters
        for character in characters:
            if character.name == "Albedo" and character.talent is not None:
                found_talent_albedo = True
                break
        if not found_talent_albedo:
            # not found
            return value
        if (
            value.cost.label & CostLabels.NORMAL_ATTACK.value > 0
            and match.player_tables[self.position.player_idx].plunge_satisfied
        ):
            # plunging attack
            assert mode == "REAL"
            value.damage += 1
        return value


# Skills


class AbiogenesisSolarIsotoma(ElementalSkillBase):
    name: Literal["Abiogenesis: Solar Isotoma"] = "Abiogenesis: Solar Isotoma"
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(elemental_dice_color=DieColor.GEO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        create object
        """
        return [
            self.create_summon("Solar Isotoma"),
            self.charge_self(1),
        ]


class RiteOfProgenitureTectonicTide(ElementalBurstBase):
    name: Literal[
        "Rite of Progeniture: Tectonic Tide"
    ] = "Rite of Progeniture: Tectonic Tide"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.GEO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Check if Solar Isotoma exists. If so, change self.damage to 8.
        """
        summons = match.player_tables[self.position.player_idx].summons
        isotoma_exist: bool = False
        for s in summons:
            if s.name == "Solar Isotoma":
                isotoma_exist = True
                break
        if isotoma_exist:
            # temporary change damage to 6
            self.damage = 6
        ret = super().get_actions(match)
        self.damage = 4
        return ret


# Talents


class DescentOfDivinity_4_0(SkillTalent):
    name: Literal["Descent of Divinity"]
    version: Literal["4.0"] = "4.0"
    character_name: Literal["Albedo"] = "Albedo"
    cost: Cost = Cost(elemental_dice_color=DieColor.GEO, elemental_dice_number=3)
    skill: Literal["Abiogenesis: Solar Isotoma"] = "Abiogenesis: Solar Isotoma"


# character base


class Albedo_4_0(CharacterBase):
    name: Literal["Albedo"]
    version: Literal["4.0"] = "4.0"
    element: ElementType = ElementType.GEO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase
        | AbiogenesisSolarIsotoma
        | RiteOfProgenitureTectonicTide
    ] = []
    faction: List[FactionType] = [FactionType.MONDSTADT]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Favonius Bladework - Weiss",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            AbiogenesisSolarIsotoma(),
            RiteOfProgenitureTectonicTide(),
        ]


register_class(Albedo_4_0 | DescentOfDivinity_4_0 | SolarIsotoma_4_0)



from lpsim.agents.interaction_agent import InteractionAgent
from lpsim.agents.nothing_agent import NothingAgent
from lpsim.server.deck import Deck
from lpsim.server.match import Match, MatchState
from tests.utils_for_test import (
    check_hp,
    get_random_state,
    get_test_id_from_command,
    make_respond,
    set_16_omni,
)


def test_wind_and_freedom():
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "sw_char 1 0",
            "TEST 1 opposite status",
            "skill 2 0 1 2 3 4",
            "choose 0",
            "card 0 0 0",
            "TEST 11 record opposite command number",
            "skill 1 0 1 2",
            "TEST 2 number -2",
            "card 0 0 0",
            "skill 1 0 1 2",
            "TEST 2 number -3",
            "sw_char 3 0",
            "end",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "card 0 0 0",
            "card 0 0 0",
            "TEST 1 one team status",
            "skill 0 0 1 2",
            "choose 1",
            "TEST 1 one team status",
            "TEST 11 record opposite command number",
            "skill 1 0 1 2",
            "TEST 2 status none, number -1",
            "sw_char 2 0",
            "TEST 3 number changed",
            "choose 1",
            "choose 3",
            "end",
        ],
        only_use_command=True,
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        default_version:4.0
        # character:Fischl
        # character:Mona
        character:Nahida*10
        # Gambler's Earrings*2
        # Wine-Stained Tricorne*2
        # Vanarana
        # Timmie*2
        # Rana*2
        # Covenant of Rock
        # Wind and Freedom
        # The Bestest Travel Companion!*2
        # Strategize*2
        Wind and Freedom*30
        """
    )
    for character in deck.characters:
        character.hp = 2
        character.max_hp = 2
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.character_number = 10
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    match.start()
    match.step()

    cmd_record_0 = 0
    cmd_record_1 = 0
    while True:
        if match.need_respond(0):
            while True:
                cmd = agent_0.commands[0]
                test_id = get_test_id_from_command(agent_0)
                if test_id == 1:
                    assert len(match.player_tables[1].team_status) == 1
                    assert match.player_tables[1].team_status[0].usage == 1
                elif test_id == 11:
                    cmd_record_1 = len(agent_1.commands)
                elif test_id == 2:
                    assert cmd_record_1 + int(cmd.strip().split()[-1]) == len(
                        agent_1.commands
                    )
                else:
                    break
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                cmd = agent_1.commands[0]
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    assert len(match.player_tables[1].team_status) == 1
                    assert match.player_tables[1].team_status[0].usage == 1
                elif test_id == 11:
                    cmd_record_0 = len(agent_0.commands)
                elif test_id == 2:
                    assert len(match.player_tables[1].team_status) == 0
                    assert cmd_record_0 + int(cmd.strip().split()[-1]) == len(
                        agent_0.commands
                    )
                elif test_id == 3:
                    assert cmd_record_0 != len(agent_0.commands) + 1
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert match.state != MatchState.ERROR


def test_wind_and_freedom_one_round():
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "card 0 0 0",
            "card 0 0 0",
            "TEST 1 one team status",
            "end",
            "TEST 2 no status",
            "end",
        ],
        only_use_command=True,
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        default_version:4.0
        # character:Fischl
        # character:Mona
        character:Nahida*10
        # Gambler's Earrings*2
        # Wine-Stained Tricorne*2
        # Vanarana
        # Timmie*2
        # Rana*2
        # Covenant of Rock
        # Wind and Freedom
        # The Bestest Travel Companion!*2
        # Strategize*2
        Wind and Freedom*30
        """
    )
    for character in deck.characters:
        character.hp = 2
        character.max_hp = 2
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.character_number = 10
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    assert len(match.player_tables[1].team_status) == 1
                    assert match.player_tables[1].team_status[0].usage == 1
                elif test_id == 2:
                    assert len(match.player_tables[1].team_status) == 0
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert match.state != MatchState.ERROR


def test_stone_thunder_nature():
    cmd_records = [
        [
            "sw_card 0 2",
            "choose 0",
            "reroll",
            "card 2 0 4",
            "sw_card 2 0",
            "TEST 1 p0 no stone in hand",
            "TEST 1 p1 no stone in hand",
            "TEST 1 p1 no thunder in hand",
            "card 0 0",
            "sw_char 1 6",
            "end",
            "reroll",
            "card 4 0 7",
            "sw_card 0 2 1 3 5",
            "card 4 0 1",
            "sw_card 3 5",
            "TEST 1 p0 no nature card",
            "TEST 1 p0 no thunder card",
            "card 4 0 5 4 3",
            "card 2 0 2 1 0",
            "end",
            "reroll",
            "TEST 2 p0 11 dice",
            "TEST 2 p1 8 dice",
            "end",
        ],
        [
            "sw_card 0 2 3 4",
            "choose 0",
            "reroll",
            "card 1 0 2",
            "sw_card 0 1 2 3",
            "TEST 3 p0 all cryo",
            "card 2 0",
            "TEST 3 p1 all omni",
            "card 2 0 6 5 4",
            "end",
            "reroll",
            "TEST 2 p1 11 dice",
            "TEST 4 p1 no team status",
            "sw_char 1 10",
            "end",
            "reroll",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Ganyu
        character:Fischl
        character:Keqing
        Stone and Contracts*7
        Thunder and Eternity*7
        Thunder and Eternity@3.7*7
        Nature and Wisdom*7
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                name = cmd[4]
                for card in match.player_tables[pidx].hands:
                    assert name not in card.name.lower()
            elif test_id == 2:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                dnum = int(cmd[3])
                assert len(match.player_tables[pidx].dice.colors) == dnum
            elif test_id == 3:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                color = cmd[4].upper()
                colors = match.player_tables[pidx].dice.colors
                assert colors == [color] * len(colors)
            elif test_id == 4:
                assert len(match.player_tables[1].team_status) == 0
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_fatui_monster():
    cmd_records = [
        [
            "sw_card 1 0",
            "choose 0",
            "end",
            "end",
            "TEST 2 card 1 cannot use",
            "skill 1 15 14 13",
            "card 3 0 12 11",
            "card 7 0 10 9",
            "card 4 0 8 7",
            "TEST 2 card 3 cannot use",
            "card 0 0 6 5",
            "card 1 0 4 3",
            "card 2 0 2 1",
            "end",
            "TEST 1 2 10 10 8 5 4",
            "skill 1 15 14 13",
            "TEST 1 2 10 10 6 5 0",
            "TEST 3 p1c0 status 1",
            "card 0 0 12 11 10 9",
            "TEST 4 no summon",
            "end",
            "skill 0 15 14 13",
            "sw_char 1 12",
            "skill 0 11 10 9",
            "skill 0 8 7 6",
        ],
        [
            "sw_card",
            "choose 0",
            "end",
            "end",
            "skill 1 15 14 13",
            "sw_char 1 12",
            "skill 0 11 10 9",
            "TEST 1 8 10 10 9 6 10",
            "TEST 5 p1 usage 1 1 1 1",
            "sw_char 2 8",
            "skill 0 7 6 5",
            "TEST 1 5 10 10 9 6 10",
            "end",
            "card 4 0 15 14",
            "skill 0 13 12 11",
            "choose 0",
            "end",
            "sw_char 1 15",
            "sw_char 0 14",
            "card 0 0 13 12",
            "card 0 0 11 10",
            "card 0 0 9 8",
            "card 1 0 7 6",
            "TEST 5 p0 usage 1 2 2 2 2",
            "sw_char 1 5",
            "TEST 5 p0 usage 1 1 1 1 1",
            "TEST 1 2 6 7 3 4 0",
            "TEST 3 p0c2 status 1",
            "TEST 2 card 1 cannot use",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Mona
        character:Nahida
        character:Klee
        Fatui Conspiracy*10
        Abyssal Summons*10
        Guardian's Oath*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                cmd = cmd.split()
                card_num = int(cmd[3])
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        assert req.card_idx != card_num
            elif test_id == 3:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                cidx = int(cmd[2][3])
                usages = [int(x) for x in cmd[4:]]
                status = match.player_tables[pidx].characters[cidx].status
                assert len(status) == len(usages)
                for s, u in zip(status, usages):
                    assert s.usage == u
            elif test_id == 4:
                for table in match.player_tables:
                    assert len(table.summons) == 0
            elif test_id == 5:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                usages = [int(x) for x in cmd[4:]]
                status = match.player_tables[pidx].team_status
                assert len(status) == len(usages)
                for s, u in zip(status, usages):
                    assert s.usage == u
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    test_wind_and_freedom()
    test_stone_thunder_nature()
    test_fatui_monster()



from typing import Dict, List, Literal

from ....struct import ObjectPosition

from ....action import ChargeAction, CreateRandomObjectAction

from .....utils.desc_registry import DescDictType

from .....utils.class_registry import register_class

from ....character.character_base import ElementalNormalAttackBase

from ....summon.base import DefendSummonBase

from ....consts import ELEMENT_TO_DAMAGE_TYPE, DamageElementalType, ObjectPositionType

from ....match import Match
from ....character.hydro.rhodeia_4_2 import (
    RhodeiaOfLoch_4_2,
    RhodeiaElementSkill as RES_4_2,
    TideAndTorrent,
)


class Frog_4_3(DefendSummonBase):
    name: Literal["Oceanic Mimic: Frog"] = "Oceanic Mimic: Frog"
    version: Literal["4.3"] = "4.3"
    usage: int = 1
    max_usage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    damage: int = 2
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 1
    attack_until_run_out_of_usage: bool = True


mimic_names = [
    "Oceanic Mimic: Squirrel",
    "Oceanic Mimic: Frog",
    "Oceanic Mimic: Raptor",
]


class RhodeiaElementSkill(RES_4_2):
    version: Literal["4.3"] = "4.3"

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.name == "Oceanid Mimic Summoning":
            pass
        else:
            assert self.name == "The Myriad Wilds"
            self._summon_number = 2
            self.cost.elemental_dice_number = 5

    def get_actions(
        self, match: Match
    ) -> List[ChargeAction | CreateRandomObjectAction]:
        """
        create object
        """
        exist_names, unexist_names = self._get_exist_unexist_names(match)
        ret: List[ChargeAction | CreateRandomObjectAction] = []
        target_position = ObjectPosition(
            player_idx=self.position.player_idx, area=ObjectPositionType.SUMMON, id=-1
        )
        if len(exist_names) == 2:
            # already 2, only refresh existing
            ret.append(
                CreateRandomObjectAction(
                    object_names=exist_names,
                    object_position=target_position,
                    object_arguments={"version": self.version},
                    number=self._summon_number,
                )
            )
        elif (
            len(unexist_names) >= self._summon_number
            and self._summon_number + len(exist_names) <= 2
        ):
            # enough unexist and will not exceed 2
            ret.append(
                CreateRandomObjectAction(
                    object_names=unexist_names,
                    object_position=target_position,
                    object_arguments={"version": self.version},
                    number=self._summon_number,
                )
            )
        else:
            # not enough unexist, create all - 1 unexist and refresh random exist
            ret.append(
                CreateRandomObjectAction(
                    object_names=unexist_names,
                    object_position=target_position,
                    object_arguments={"version": self.version},
                    number=len(unexist_names) - 1,
                )
            )
            ret.append(
                CreateRandomObjectAction(
                    object_names=exist_names,
                    object_position=target_position,
                    object_arguments={"version": self.version},
                    number=self._summon_number - len(unexist_names) + 1,
                )
            )
        ret.append(self.charge_self(1))
        return ret


class RhodeiaOfLoch_4_3(RhodeiaOfLoch_4_2):
    version: Literal["4.3"] = "4.3"
    skills: List[ElementalNormalAttackBase | RhodeiaElementSkill | TideAndTorrent] = []

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="Surge",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            RhodeiaElementSkill(
                name="Oceanid Mimic Summoning",
            ),
            RhodeiaElementSkill(
                name="The Myriad Wilds",
            ),
            TideAndTorrent(),
        ]


desc: Dict[str, DescDictType] = {
    "CHARACTER/Rhodeia of Loch": {
        "descs": {"4.3": {"zh-CN": "", "en-US": ""}},
    },
    "SKILL_Rhodeia of Loch_ELEMENTAL_BURST/Tide and Torrent": {
        "descs": {
            "4.3": {
                "zh-CN": "$SKILL_Rhodeia of Loch_ELEMENTAL_BURST/Tide and Torrent|descs|4.2|zh-CN",  # noqa: E501
                "en-US": "$SKILL_Rhodeia of Loch_ELEMENTAL_BURST/Tide and Torrent|descs|4.2|en-US",  # noqa: E501
            }
        }
    },
    "SKILL_Rhodeia of Loch_ELEMENTAL_SKILL/Oceanid Mimic Summoning": {
        "descs": {
            "4.3": {
                "zh-CN": "随机召唤1种纯水幻形。（最多生成2种，超过时随机生成已存在的纯水幻形）",  # noqa: E501
                "en-US": "Randomly summons 1 Oceanid Mimic (Maximum 2 types of mimics. If exceeded, randomly generate preexisting ones.).",  # noqa: E501
            }
        }
    },
    "SKILL_Rhodeia of Loch_ELEMENTAL_SKILL/The Myriad Wilds": {
        "descs": {
            "4.3": {
                "zh-CN": "随机召唤2种纯水幻形。（最多生成2种，超过时随机生成已存在的纯水幻形）",  # noqa: E501
                "en-US": "Randomly summons 2 Oceanid Mimic (Maximum 2 types of mimics. If exceeded, randomly generate preexisting ones.).",  # noqa: E501
            }
        }
    },
    "SKILL_Rhodeia of Loch_NORMAL_ATTACK/Surge": {
        "descs": {
            "4.3": {"zh-CN": "造成1点水元素伤害。", "en-US": "Deals 1 Hydro DMG."}
        }
    },
    "SUMMON/Oceanic Mimic: Frog": {
        "descs": {
            "4.3": {
                "zh-CN": "我方出战角色受到伤害时：抵消1点伤害。\n可用次数：1，耗尽时不弃置此牌。\n\n结束阶段：如果可用次数已耗尽：弃置此牌，造成2点水元素伤害。",  # noqa: E501
                "en-US": "When your active character takes DMG: Decrease DMG taken by 1. When the Usages are depleted, this card will not be discarded. At the End Phase, if Usage(s) have been depleted: Discard this card, deal 2 Hydro DMG. Usage(s): 1",  # noqa: E501
            }
        },
    },
}


register_class(RhodeiaOfLoch_4_3 | Frog_4_3, desc)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Skills


class FantasticVoyage(ElementalBurstBase):
    name: Literal["Fantastic Voyage"] = "Fantastic Voyage"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO, elemental_dice_number=4, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(
            match,
            [
                self.create_team_status(
                    "Inspiration Field",
                    {"talent_activated": self.is_talent_equipped(match)},
                )
            ],
        )


# Talents


class GrandExpectation_3_3(SkillTalent):
    name: Literal["Grand Expectation"]
    version: Literal["3.3"] = "3.3"
    character_name: Literal["Bennett"] = "Bennett"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO, elemental_dice_number=4, charge=2
    )
    skill: Literal["Fantastic Voyage"] = "Fantastic Voyage"


# character base


class Bennett_3_3(CharacterBase):
    name: Literal["Bennett"]
    version: Literal["3.3"] = "3.3"
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[PhysicalNormalAttackBase | ElementalSkillBase | FantasticVoyage] = []
    faction: List[FactionType] = [FactionType.MONDSTADT]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Strike of Fortune",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            ElementalSkillBase(
                name="Passion Overload",
                damage_type=DamageElementalType.PYRO,
                cost=ElementalSkillBase.get_cost(self.element),
            ),
            FantasticVoyage(),
        ]


register_class(Bennett_3_3 | GrandExpectation_3_3)



import os

from lpsim.server.deck import Deck

from lpsim.server.match import Match
from lpsim.agents.interaction_agent import InteractionAgent
from lpsim import MatchState
from tests.utils_for_test import (
    check_hp,
    check_usage,
    get_pidx_cidx,
    get_random_state,
    get_test_id_from_command,
    make_respond,
    read_from_log_json,
    set_16_omni,
)


def test_azhdaha():
    match, agent_0, agent_1 = read_from_log_json(
        os.path.join(os.path.dirname(__file__), "jsons", "test_azhdaha.json")
    )
    # modify hp
    # for i in range(2):
    #     characters = match.player_tables[i].player_deck_information.characters  # noqa: E501
    #     for c in characters:
    #         c.hp = c.max_hp = 30
    # add omnipotent guide
    # set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                pidx = int(cmd[2][1])
                colors = {}
                for c in cmd[3:]:
                    c, n = c.strip().split("*")
                    colors[c] = int(n)
                for c in match.player_tables[pidx].dice.colors:
                    c = c.value.lower()
                    assert c in colors
                    colors[c] -= 1
                    if colors[c] == 0:
                        del colors[c]
            elif test_id == 3:
                pidx, cidx = get_pidx_cidx(cmd)
                status = match.player_tables[pidx].characters[cidx].status
                check_usage(status, cmd[4:])
            elif test_id == 4:
                pidx, cidx = get_pidx_cidx(cmd)
                desc = cmd[-1]
                assert match.player_tables[pidx].characters[cidx].desc.lower() == desc
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_azhdaha_2():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "sw_char 2 15",
            "TEST 4 p0c2 desc empty",
            "skill 1 14 13 12",
            "TEST 4 p0c2 desc pyro",
            "sw_char 1 12",
            "skill 1 11 10 9",
            "TEST 4 p0c1 desc cryo",
            "end",
        ],
        [
            "sw_card",
            "choose 4",
            "skill 0 15 14 13",
            "skill 0 12 11 10",
            "sw_char 3 9",
            "skill 1 8 7 6",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.3
        character:Azhdaha*3
        character:Kaeya
        character:Klee
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 4:
                pidx, cidx = get_pidx_cidx(cmd)
                desc = cmd[-1]
                if desc == "empty":
                    desc = ""
                assert match.player_tables[pidx].characters[cidx].desc.lower() == desc
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    test_azhdaha()
    test_azhdaha_2()



from .interaction_agent import InteractionAgent
from .nothing_agent import NothingAgent
from .random_agent import RandomAgent
from .agent_base import AgentBase  # noqa: F401


Agents = InteractionAgent | NothingAgent | RandomAgent



"""
Arcane legend cards.
"""

from typing import Any, List, Literal

from ...event import UseCardEventArguments

from ....utils.class_registry import register_class

from ...modifiable_values import DamageValue

from ...action import (
    Actions,
    ConsumeArcaneLegendAction,
    CreateDiceAction,
    CreateObjectAction,
    DrawCardAction,
    MakeDamageAction,
    RemoveCardAction,
)

from ...struct import Cost, MultipleObjectPosition, ObjectPosition
from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE,
    CostLabels,
    DamageType,
    ElementType,
    ObjectPositionType,
    ObjectType,
)
from ...object_base import EventCardBase


class ArcaneLegendBase(EventCardBase):
    type: Literal[ObjectType.ARCANE] = ObjectType.ARCANE
    cost_label: int = (
        CostLabels.CARD.value | CostLabels.ARCANE.value | CostLabels.EVENT.value
    )

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ConsumeArcaneLegendAction]:
        """
        Consume arcane legend.
        """
        assert target is None
        return [ConsumeArcaneLegendAction(player_idx=self.position.player_idx)]

    def event_handler_USE_CARD(
        self, event: UseCardEventArguments, match: Any
    ) -> List[Actions]:
        """
        Mainly copied from CardBase with modifications with use_card_success.

        If this card is used, trigger events. But if not use_card_success,
        instead of do nothing, will consume arcane legend.
        """
        actions: List[Actions] = []
        if event.action.card_position.id != self.id:
            # not this card
            return actions
        if self.remove_when_used:  # pragma: no branch
            actions.append(
                RemoveCardAction(
                    position=ObjectPosition(
                        player_idx=self.position.player_idx,
                        area=ObjectPositionType.HAND,
                        id=self.id,
                    ),
                    remove_type="USED",
                )
            )
        # target is None, otherwise should match class
        assert event.action.target is None or (
            isinstance(event.action.target, MultipleObjectPosition)
            == self.target_position_type_multiple
        )
        if event.use_card_success:
            # use success, add actions of the card
            actions += self.get_actions(
                target=event.action.target,  # type: ignore
                match=match,
            )
        else:
            # use failed, only consume arcane legend
            actions += [ConsumeArcaneLegendAction(player_idx=self.position.player_idx)]
        return actions


class AncientCourtyard_3_8(ArcaneLegendBase):
    name: Literal["Ancient Courtyard"]
    version: Literal["3.8"] = "3.8"
    cost: Cost = Cost(arcane_legend=True)

    def is_valid(self, match: Any) -> bool:
        """
        You must have a character who has already equipped a Weapon or Artifact
        """
        for character in match.player_tables[self.position.player_idx].characters:
            if character.weapon is not None or character.artifact is not None:
                return True
        return False

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        No targets.
        """
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ConsumeArcaneLegendAction | CreateObjectAction]:
        """
        create team status
        """
        assert target is None
        ret: List[ConsumeArcaneLegendAction | CreateObjectAction] = []
        ret += super().get_actions(target, match)
        ret.append(
            CreateObjectAction(
                object_name=self.name,
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectType.TEAM_STATUS,
                    id=0,
                ),
                object_arguments={},
            )
        )
        return ret


class CovenantOfRock_3_8(ArcaneLegendBase):
    name: Literal["Covenant of Rock"]
    version: Literal["3.8"] = "3.8"
    cost: Cost = Cost(arcane_legend=True)

    def is_valid(self, match: Any) -> bool:
        """
        Can only be played when you have 0 Elemental Dice left.
        """
        return len(match.player_tables[self.position.player_idx].dice.colors) == 0

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        No targets.
        """
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ConsumeArcaneLegendAction | CreateDiceAction]:
        """
        Generate 2 different Elemental Dice.
        """
        assert target is None
        ret: List[ConsumeArcaneLegendAction | CreateDiceAction] = []
        ret += super().get_actions(target, match)
        ret.append(
            CreateDiceAction(
                player_idx=self.position.player_idx, number=2, different=True
            )
        )
        return ret


class JoyousCelebration_4_2(ArcaneLegendBase):
    name: Literal["Joyous Celebration"]
    version: Literal["4.2"] = "4.2"
    cost: Cost = Cost(same_dice_number=1, arcane_legend=True)
    apply_no_element_character: bool = False

    def is_valid(self, match: Any) -> bool:
        """
        Your active character must be one of the following elemental types to
        play this card: Cryo/Hydro/Pyro/Electro/Dendro
        """
        character = match.player_tables[self.position.player_idx].get_active_character()
        return character.element in [
            ElementType.CRYO,
            ElementType.HYDRO,
            ElementType.PYRO,
            ElementType.ELECTRO,
            ElementType.DENDRO,
        ]

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        No targets.
        """
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ConsumeArcaneLegendAction | MakeDamageAction]:
        """
        apply element to corresponding characters.
        """
        assert target is None
        ret: List[ConsumeArcaneLegendAction | MakeDamageAction] = []
        ret += super().get_actions(target, match)
        damage_action = MakeDamageAction(
            damage_value_list=[],
        )
        active_character = match.player_tables[
            self.position.player_idx
        ].get_active_character()
        element = active_character.element
        for c in match.player_tables[self.position.player_idx].characters:
            if c.is_defeated:
                continue
            if (not self.apply_no_element_character) and len(
                c.element_application
            ) == 0:
                # not apply to no element character
                continue
            damage_action.damage_value_list.append(
                DamageValue(
                    position=self.position,
                    damage_type=DamageType.ELEMENT_APPLICATION,
                    target_position=c.position,
                    damage=0,
                    damage_elemental_type=ELEMENT_TO_DAMAGE_TYPE[element],
                    cost=self.cost.copy(),
                )
            )
        if len(damage_action.damage_value_list) > 0:
            ret.append(damage_action)
        return ret


class JoyousCelebration_4_0(JoyousCelebration_4_2):
    version: Literal["4.0"] = "4.0"
    apply_no_element_character: bool = True


class FreshWindOfFreedom_4_1(ArcaneLegendBase):
    name: Literal["Fresh Wind of Freedom"]
    version: Literal["4.1"] = "4.1"
    cost: Cost = Cost(arcane_legend=True)

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ConsumeArcaneLegendAction | CreateObjectAction]:
        """
        Create Wind and Freedom team status.
        """
        assert target is None
        return super().get_actions(target, match) + [
            CreateObjectAction(
                object_name=self.name,
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=-1,
                ),
                object_arguments={},
            )
        ]


class InEveryHouseAStove_4_2(ArcaneLegendBase):
    name: Literal["In Every House a Stove"]
    version: Literal["4.2"] = "4.2"
    cost: Cost = Cost(arcane_legend=True)

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ConsumeArcaneLegendAction | DrawCardAction]:
        """
        Draw cards.
        """
        assert target is None
        return super().get_actions(target, match) + [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=min(match.round_number, 4),
                draw_if_filtered_not_enough=True,
            )
        ]


register_class(
    AncientCourtyard_3_8
    | CovenantOfRock_3_8
    | JoyousCelebration_4_2
    | FreshWindOfFreedom_4_1
    | InEveryHouseAStove_4_2
    | JoyousCelebration_4_0
)



from typing import Literal, List

from ....utils.class_registry import register_class

from ..character_base import PhysicalNormalAttackBase
from ...consts import DieColor
from ...struct import Cost
from .ganyu_3_7 import Ganyu_3_7 as G_3_7
from .ganyu_3_7 import UndividedHeart_3_7 as UH_3_7
from .ganyu_3_7 import CelestialShower as CS_3_7
from .ganyu_3_7 import FrostflakeArrow, TrailOftheQilin


class CelestialShower(CS_3_7):
    damage: int = 1
    cost: Cost = Cost(
        elemental_dice_color=DieColor.CRYO, elemental_dice_number=3, charge=2
    )


class UndividedHeart_3_3(UH_3_7):
    version: Literal["3.3"] = "3.3"


class Ganyu_3_3(G_3_7):
    version: Literal["3.3"] = "3.3"
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | TrailOftheQilin | FrostflakeArrow | CelestialShower
    ] = []

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Liutian Archery",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            TrailOftheQilin(),
            FrostflakeArrow(),
            CelestialShower(),
        ]


register_class(Ganyu_3_3 | UndividedHeart_3_3)



from typing import Any, Dict, Literal, List, Tuple

from ...object_base import CreateSystemEventHandlerObject

from ....utils.class_registry import register_base_class, register_class

from ...modifiable_values import CombatActionValue, CostValue, FullCostValue

from ...dice import Dice

from .base import (
    LimitedEffectSupportBase,
    RoundEffectSupportBase,
    SupportBase,
    UsageWithRoundRestrictionSupportBase,
)
from ...consts import (
    DIE_COLOR_TO_ELEMENT,
    ELEMENT_DEFAULT_ORDER,
    CostLabels,
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    ELEMENT_TO_DIE_COLOR,
    ElementalReactionType,
    IconType,
    ObjectPositionType,
    PlayerActionLabels,
    SkillType,
)
from ...struct import Cost, ObjectPosition
from ...action import (
    Actions,
    ChangeObjectUsageAction,
    ChargeAction,
    CreateDiceAction,
    DrawCardAction,
    MoveObjectAction,
    RemoveDiceAction,
    RemoveObjectAction,
)
from ...event import (
    ActionEndEventArguments,
    ChangeObjectUsageEventArguments,
    ChooseCharacterEventArguments,
    MoveObjectEventArguments,
    PlayerActionStartEventArguments,
    ReceiveDamageEventArguments,
    RemoveObjectEventArguments,
    RoundEndEventArguments,
    RoundPrepareEventArguments,
    SkillEndEventArguments,
    SwitchCharacterEventArguments,
    UseCardEventArguments,
)


class CompanionBase(SupportBase):
    cost_label: int = CostLabels.CARD.value | CostLabels.COMPANION.value


register_base_class(CompanionBase)


class RoundEffectCompanionBase(RoundEffectSupportBase, CompanionBase):
    cost_label: int = CostLabels.CARD.value | CostLabels.COMPANION.value


class Paimon_3_3(CompanionBase):
    name: Literal["Paimon"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(same_dice_number=3)
    usage: int = 2
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[CreateDiceAction | RemoveObjectAction]:
        """
        When in round prepare, increase usage.
        If usage is 3, remove self, draw a card and create a dice.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        self.usage -= 1
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx,
                number=2,
                color=DieColor.OMNI,
            )
        ] + self.check_should_remove()


class Katheryne_3_6(RoundEffectCompanionBase):
    name: Literal["Katheryne"]
    version: Literal["3.6"] = "3.6"
    cost: Cost = Cost(same_dice_number=1)
    max_usage_per_round: int = 1

    def value_modifier_COMBAT_ACTION(
        self, value: CombatActionValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CombatActionValue:
        """
        Copied from team status Leave It to Me!
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return value
        if self.usage <= 0:
            # no usage
            return value
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
        ):
            # not self switch character, do nothing
            return value
        if value.action_label & PlayerActionLabels.SWITCH.value == 0:
            # not switch character, do nothing
            return value
        if not value.do_combat_action:
            # already quick action, do nothing
            return value
        # self switch character, change to quick action
        value.do_combat_action = False
        assert mode == "REAL"
        self.usage -= 1
        return value


class Katheryne_3_3(Katheryne_3_6):
    version: Literal["3.3"]
    cost: Cost = Cost(any_dice_number=2)


class Timaeus_3_3(CompanionBase, UsageWithRoundRestrictionSupportBase):
    name: Literal["Timaeus"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(same_dice_number=2)
    max_usage_one_round: int = 1
    decrease_target: int = CostLabels.ARTIFACT.value
    usage: int = 2
    icon_type: Literal[IconType.COUNTER] = IconType.COUNTER

    def play(self, match: Any) -> List[Actions]:
        UsageWithRoundRestrictionSupportBase.play(self, match)
        self.usage = 2
        return []

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[Actions]:
        """
        When in round end, increase usage.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        self.usage += 1
        return []

    def value_modifier_FULL_COST(
        self, value: FullCostValue, match: Any, mode: Literal["REAL", "TEST"]
    ) -> FullCostValue:
        """
        When in support, and self equip, and usage is enough, and has round
        usage, decrease cost.
        """
        decrease_number = value.cost.total_dice_cost
        if (
            self.position.area != ObjectPositionType.SUPPORT
            or self.position.player_idx != value.position.player_idx
            or value.cost.label & self.decrease_target == 0
            or decrease_number > self.usage
            or decrease_number == 0
            or self.usage_this_round <= 0
        ):
            # not in support, not self player, not target equip, not enough
            # usage this round or total usage, do nothing
            return value
        # decrease all cost
        for _ in range(decrease_number):
            value.cost.decrease_cost(None)
        if mode == "REAL":
            self.usage_this_round -= 1
            self.usage -= decrease_number
        return value


class Wagner_3_3(Timaeus_3_3):
    name: Literal["Wagner"]
    decrease_target: int = CostLabels.WEAPON.value


class ChefMao_4_1(RoundEffectCompanionBase, LimitedEffectSupportBase):
    name: Literal["Chef Mao"]
    version: Literal["4.1"] = "4.1"
    cost: Cost = Cost(same_dice_number=1)
    max_usage_per_round: int = 1
    limited_usage: int = 1

    def _limited_action(self, match: Any) -> List[DrawCardAction]:
        return [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=1,
                whitelist_cost_labels=CostLabels.FOOD.value,
                draw_if_filtered_not_enough=False,
            )
        ]

    def event_handler_USE_CARD(
        self, event: UseCardEventArguments, match: Any
    ) -> List[Actions]:
        """
        When use food card, create one random elemental die and do limited
        action.
        """
        if self.position.area == ObjectPositionType.SUPPORT:
            # on support area, do effect
            if self.usage <= 0:
                # usage is 0, do nothing
                return []
            if event.action.card_position.player_idx != self.position.player_idx:
                # not our character use card, do nothing
                return []
            if event.card_cost.label & CostLabels.FOOD.value == 0:
                # not food card, do nothing
                return []
            # our use food card, generate die and do limited action
            self.usage -= 1
            return [
                CreateDiceAction(
                    player_idx=self.position.player_idx,
                    number=1,
                    different=True,  # use different to avoid generating OMNI
                )
            ] + self.do_limited_action(match)
        # otherwise, do normal response
        return super().event_handler_USE_CARD(event, match)


class ChefMao_3_3(ChefMao_4_1):
    version: Literal["3.3"]
    limited_usage: int = 0


class Tubby_3_3(RoundEffectCompanionBase):
    name: Literal["Tubby"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(same_dice_number=2)
    max_usage_per_round: int = 1
    decrease_target: int = CostLabels.LOCATION.value
    decrease_cost: int = 2

    def use(self, mode: Literal["REAL", "TEST"], result: List[bool]) -> None:
        """
        When used, check whether need to decrease usage.
        """
        if True in result:
            if mode == "REAL":
                self.usage -= 1

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["REAL", "TEST"]
    ) -> CostValue:
        """
        If in support and self use target card, decrease cost by 2.
        """
        if self.usage <= 0:
            # no usage
            return value
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return value
        if value.position.player_idx != self.position.player_idx:
            # not self player
            return value
        if value.cost.label & self.decrease_target == 0:
            # not decrease target
            return value
        # decrease
        result = [value.cost.decrease_cost(None) for _ in range(self.decrease_cost)]
        self.use(mode, result)
        return value


class Timmie_3_3(CompanionBase):
    name: Literal["Timmie"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost()
    usage: int = 0
    max_usage: int = 3
    icon_type: Literal[IconType.COUNTER] = IconType.COUNTER

    def play(self, match: Any) -> List[ChangeObjectUsageAction]:
        """
        When played, first reset usage to 0, then increase usage.
        """
        self.usage = 0
        return [ChangeObjectUsageAction(object_position=self.position, change_usage=1)]

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        When in round prepare, increase usage.
        If usage is 3, remove self, draw a card and create a dice.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        return [ChangeObjectUsageAction(object_position=self.position, change_usage=1)]

    def event_handler_CHANGE_OBJECT_USAGE(
        self, event: ChangeObjectUsageEventArguments, match: Any
    ) -> List[RemoveObjectAction | DrawCardAction | CreateDiceAction]:
        """
        when self usage changed to 3, remove self, draw a card and create a
        dice.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        if event.action.object_position.id != self.id:
            # not self
            return []
        ret: List[RemoveObjectAction | DrawCardAction | CreateDiceAction] = []
        if self.usage == 3:
            ret += [
                RemoveObjectAction(
                    object_position=self.position,
                ),
                DrawCardAction(
                    player_idx=self.position.player_idx,
                    number=1,
                    draw_if_filtered_not_enough=True,
                ),
                CreateDiceAction(
                    player_idx=self.position.player_idx,
                    number=1,
                    color=DieColor.OMNI,
                ),
            ]
        return ret


class Liben_3_3(CompanionBase):
    name: Literal["Liben"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost()
    usage: int = 0
    max_usage: int = 3
    icon_type: Literal[IconType.COUNTER] = IconType.COUNTER

    def play(self, match: Any) -> List[Actions]:
        self.usage = 0
        return []

    def _sort_colors(self, colors: List[DieColor]) -> List[Tuple[int, int, DieColor]]:
        """
        Sort colors. Omni always last, others has two keys, 1. number of
        this color, 2. default order.
        Return: List[Tuple[number, default_order, color]]. For Omni, the
        default order is filled with 0.
        """
        colors_dict: Dict[DieColor, int] = {DieColor.OMNI: 0}
        for element in ELEMENT_DEFAULT_ORDER:
            colors_dict[ELEMENT_TO_DIE_COLOR[element]] = 0
        for color in colors:
            colors_dict[color] += 1
        # first: number, second: default order, third: color
        colors_lists: List[Tuple[int, int, DieColor]] = []
        # specially handle OMNI
        omni_num = 0
        for color in colors_dict:
            if colors_dict[color] == 0:
                continue
            if color == DieColor.OMNI:
                omni_num = colors_dict[color]
                continue
            colors_lists.append(
                (
                    colors_dict[color],
                    ELEMENT_DEFAULT_ORDER.index(DIE_COLOR_TO_ELEMENT[color]),
                    color,
                )
            )
        # number desc, default order asc
        colors_lists.sort(key=lambda x: (-x[0], x[1]))
        if omni_num > 0:
            colors_lists.append((omni_num, 0, DieColor.OMNI))
        return colors_lists

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[RemoveDiceAction]:
        """
        Collect different color dice in round end.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        collect_order: List[DieColor] = []
        dice: Dice = match.player_tables[self.position.player_idx].dice
        colors_lists = self._sort_colors(dice.colors)
        omni_num = 0
        if len(colors_lists) == 0:
            # no dice, do nothing
            return []
        if colors_lists[-1][2] == DieColor.OMNI:
            omni_num = colors_lists.pop()[0]
        # generate collect order
        color_order = [x[2] for x in colors_lists] + [DieColor.OMNI] * omni_num
        # only collect dice if usage is not full
        collect_order = color_order[: self.max_usage - self.usage]
        self.usage += len(collect_order)
        return [
            RemoveDiceAction(
                player_idx=self.position.player_idx,
                dice_idxs=dice.colors_to_idx(collect_order),
            )
        ]

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[CreateDiceAction | DrawCardAction | RemoveObjectAction]:
        """
        If this card has collected 3 Elemental Dice, draw 2 cards and create
        Omni Element x2, then discard this card.'
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        if self.usage != self.max_usage:
            # not enough dice collected
            return []
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx,
                number=2,
                color=DieColor.OMNI,
            ),
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=2,
                draw_if_filtered_not_enough=True,
            ),
            RemoveObjectAction(
                object_position=self.position,
            ),
        ]


class ChangTheNinth_3_3(CompanionBase):
    name: Literal["Chang the Ninth"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost()
    usage: int = 0
    max_usage: int = 3
    inspiration_got: bool = False
    icon_type: Literal[IconType.COUNTER] = IconType.COUNTER

    def event_handler_PLAYER_ACTION_START(
        self, event: PlayerActionStartEventArguments, match: Any
    ) -> List[Actions]:
        """
        When anyone start action, reset inspiration got.
        """
        self.inspiration_got = False
        return []

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        If physical or piercing or elemental reaction, gain inspiration.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        if event.final_damage.damage_type == DamageType.HEAL:
            # heal, do nothing
            return []
        if (
            event.final_damage.damage_elemental_type
            not in [DamageElementalType.PHYSICAL, DamageElementalType.PIERCING]
            and event.final_damage.element_reaction == ElementalReactionType.NONE
        ):
            # not physical piercing and not elemental reaction, do nothing
            return []
        # get inspiration
        self.inspiration_got = True
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[DrawCardAction | RemoveObjectAction]:
        """
        If inspiration got, increase usage. If usage receives max_usage,
        remove self and draw two cards.
        """
        if self.inspiration_got:
            self.usage += 1
        self.inspiration_got = False
        if self.usage == self.max_usage:
            # remove self and draw two cards
            return [
                RemoveObjectAction(
                    object_position=self.position,
                ),
                DrawCardAction(
                    player_idx=self.position.player_idx,
                    number=2,
                    draw_if_filtered_not_enough=True,
                ),
            ]
        return []


class Ellin_3_3(CreateSystemEventHandlerObject, RoundEffectCompanionBase):
    name: Literal["Ellin"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(same_dice_number=2)
    usage: int = 0
    max_usage_per_round: int = 1

    handler_name: str = "Ellin"

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["REAL", "TEST"]
    ) -> CostValue:
        """
        When a skill id is recorded, and has usage, decrease cost.
        """
        if (
            self.position.area != ObjectPositionType.SUPPORT
            or self.usage <= 0
            or value.position.player_idx != self.position.player_idx
            or value.cost.label
            & (
                CostLabels.NORMAL_ATTACK.value
                | CostLabels.ELEMENTAL_SKILL.value
                | CostLabels.ELEMENTAL_BURST.value
            )
            == 0
        ):
            # no usage, not self player or not skill
            return value
        ellin_event_handler = self._get_event_handler(match)
        if value.position.id not in ellin_event_handler.recorded_skill_ids:
            # skill id not recorded, do nothing
            return value
        # decrease
        if value.cost.decrease_cost(value.cost.elemental_dice_color):
            if mode == "REAL":
                self.usage -= 1
        return value


class IronTongueTian_3_3(CompanionBase):
    name: Literal["Iron Tongue Tian"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(any_dice_number=2)
    usage: int = 2
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[ChargeAction | RemoveObjectAction]:
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        table = match.player_tables[self.position.player_idx]
        characters = [table.get_active_character()]
        for c in match.player_tables[self.position.player_idx].characters:
            if c.position.id != characters[0].position.id:
                characters.append(c)
        for c in characters:
            if c.is_alive and c.charge < c.max_charge:
                assert self.usage > 0
                self.usage -= 1
                return [
                    ChargeAction(
                        player_idx=c.position.player_idx,
                        character_idx=c.position.character_idx,
                        charge=1,
                    )
                ] + self.check_should_remove()
        return []


class LiuSu_3_3(CompanionBase, UsageWithRoundRestrictionSupportBase):
    name: Literal["Liu Su"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(same_dice_number=1)
    usage: int = 2
    max_usage_one_round: int = 1
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def charge(self, character: Any) -> List[ChargeAction | RemoveObjectAction]:
        """
        If this character is our character and has no energy, charge it.
        """
        if (
            self.position.area == ObjectPositionType.SUPPORT
            and character.position.player_idx == self.position.player_idx
            and character.charge == 0
            and self.has_usage()
        ):
            # in support, and our character, and no energy
            self.use()
            return [
                ChargeAction(
                    player_idx=character.position.player_idx,
                    character_idx=character.position.character_idx,
                    charge=1,
                )
            ] + self.check_should_remove()
        return []

    def event_handler_SWITCH_CHARACTER(
        self, event: SwitchCharacterEventArguments, match: Any
    ) -> List[ChargeAction | RemoveObjectAction]:
        character = match.player_tables[event.action.player_idx].characters[
            event.action.character_idx
        ]
        return self.charge(character)

    def event_handler_CHOOSE_CHARACTER(
        self, event: ChooseCharacterEventArguments, match: Any
    ) -> List[ChargeAction | RemoveObjectAction]:
        character = match.player_tables[event.action.player_idx].characters[
            event.action.character_idx
        ]
        return self.charge(character)


class Hanachirusato_3_7(CompanionBase):
    name: Literal["Hanachirusato"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost()
    usage: int = 0
    max_usage: int = 3
    decrease_target: int = CostLabels.WEAPON.value | CostLabels.ARTIFACT.value
    decrease_number: int = 2
    effect_triggered: bool = False
    icon_type: Literal[IconType.COUNTER] = IconType.COUNTER

    def event_handler_REMOVE_OBJECT(
        self, event: RemoveObjectEventArguments, match: Any
    ) -> List[Actions]:
        """
        When any sommon is removed, add one usage
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        if event.action.object_position.area != ObjectPositionType.SUMMON:
            # not summon, do nothing
            return []
        self.usage = min(self.usage + 1, self.max_usage)
        return []

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["REAL", "TEST"]
    ) -> CostValue:
        """
        When in support, and self use target card, decrease cost by 2.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return value
        if self.usage != self.max_usage:
            # not enough usage
            return value
        if value.position.player_idx != self.position.player_idx:
            # not self player
            return value
        if value.cost.label & self.decrease_target == 0:
            # not decrease target
            return value
        # decrease
        result = [
            value.cost.decrease_cost(None),
            value.cost.decrease_cost(None),
        ]
        if True in result:
            if mode == "REAL":
                self.effect_triggered = True
        return value

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Any
    ) -> List[Actions]:
        if self.position.area == ObjectPositionType.SUPPORT:
            # in support, check if effect triggered. If so, remove self.
            if self.effect_triggered:
                return [
                    RemoveObjectAction(
                        object_position=self.position,
                    )
                ]
        return super().event_handler_MOVE_OBJECT(event, match)


class KidKujirai_3_7(CompanionBase):
    name: Literal["Kid Kujirai"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost()
    usage: int = 0
    icon_type: Literal[IconType.NONE] = IconType.NONE

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[CreateDiceAction | MoveObjectAction | RemoveObjectAction]:
        """
        Create one OMNI for this side; and if other side has position, move
        to opposite; otherwise remove self.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        ret: List[CreateDiceAction | MoveObjectAction | RemoveObjectAction] = []
        ret.append(
            CreateDiceAction(
                player_idx=self.position.player_idx,
                number=1,
                color=DieColor.OMNI,
            )
        )
        opposite = match.player_tables[1 - self.position.player_idx].supports
        max_number = match.config.max_support_number
        if len(opposite) >= max_number:
            # opposite side support is full, remove self
            ret.append(
                RemoveObjectAction(
                    object_position=self.position,
                )
            )
        else:
            # opposite side support is not full, move to opposite
            ret.append(
                MoveObjectAction(
                    object_position=self.position,
                    target_position=ObjectPosition(
                        player_idx=1 - self.position.player_idx,
                        area=ObjectPositionType.SUPPORT,
                        id=self.position.id,
                    ),
                )
            )
        return ret


class Xudong_3_7(Tubby_3_3):
    name: Literal["Xudong"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(any_dice_number=2)
    decrease_target: int = CostLabels.FOOD.value
    decrease_cost: int = 2


class Dunyarzad_4_1(Tubby_3_3, LimitedEffectSupportBase):
    name: Literal["Dunyarzad"]
    version: Literal["4.1"] = "4.1"
    cost: Cost = Cost(same_dice_number=1)
    decrease_target: int = CostLabels.COMPANION.value
    decrease_cost: int = 1

    limited_usage: int = 1
    triggered: bool = False

    def _limited_action(self, match: Any) -> List[DrawCardAction]:
        return [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=1,
                whitelist_cost_labels=self.decrease_target,
                draw_if_filtered_not_enough=False,
            )
        ]

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Any
    ) -> List[Actions]:
        """
        Do limited action to draw a support card if needed
        """
        ret = super().event_handler_MOVE_OBJECT(event, match)
        if self.position.area == ObjectPositionType.SUPPORT:
            # on support area, do effect
            position = event.action.target_position
            if position.id == self.id:
                # self move, do nothing
                return ret
            if position.player_idx != self.position.player_idx:
                # not our character use card, do nothing
                return ret
            obj = match.get_object(position)
            if obj.cost.label & CostLabels.COMPANION.value == 0:
                # not companion card, do nothing
                return ret
            # our use companion card, do limited action
            return self.do_limited_action(match)
        # otherwise, do normal response
        return ret


class Dunyarzad_3_7(Dunyarzad_4_1):
    version: Literal["3.7"]
    limited_usage: int = 0


class Rana_3_7(RoundEffectCompanionBase):
    name: Literal["Rana"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(same_dice_number=2)
    max_usage_per_round: int = 1

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateDiceAction]:
        """
        if it is in support are, and self player used a elemental skill,
        and have usage, and have next character, generate a die with color
        of next character.
        """
        if (
            self.position.area == ObjectPositionType.SUPPORT
            and event.action.position.player_idx == self.position.player_idx
            and event.action.skill_type == SkillType.ELEMENTAL_SKILL
            and self.usage > 0
        ):
            table = match.player_tables[self.position.player_idx]
            next_idx = table.next_character_idx()
            if next_idx is not None:
                self.usage -= 1
                ele_type: ElementType = table.characters[next_idx].element
                die_color = ELEMENT_TO_DIE_COLOR[ele_type]
                return [
                    CreateDiceAction(
                        player_idx=self.position.player_idx,
                        number=1,
                        color=die_color,
                    )
                ]
        return []


class MasterZhang_3_8(RoundEffectCompanionBase):
    name: Literal["Master Zhang"]
    version: Literal["3.8"] = "3.8"
    cost: Cost = Cost(same_dice_number=1)
    max_usage_per_round: int = 1
    card_cost_label: int = CostLabels.WEAPON.value

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["REAL", "TEST"]
    ) -> CostValue:
        """
        If has usage, and self use weapon card, decrease.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return value
        if self.usage <= 0:
            # no usage
            return value
        if value.position.player_idx != self.position.player_idx:
            # not self player
            return value
        if value.cost.label & self.card_cost_label == 0:
            # not corresponding card
            return value
        # decrease
        decrease_number = 1
        table = match.player_tables[self.position.player_idx]
        for character in table.characters:
            if (
                character.weapon is not None
                and self.card_cost_label & CostLabels.WEAPON.value > 0
                or character.artifact is not None
                and self.card_cost_label & CostLabels.ARTIFACT.value > 0
                or character.talent is not None
                and self.card_cost_label & CostLabels.TALENT.value > 0
            ):
                decrease_number += 1
        result = []
        for _ in range(decrease_number):
            result.append(value.cost.decrease_cost(None))
        if True in result:
            if mode == "REAL":
                self.usage -= 1
        return value


class Setaria_4_0(CompanionBase):
    name: Literal["Setaria"]
    version: Literal["4.0"] = "4.0"
    cost: Cost = Cost(same_dice_number=1)
    usage: int = 3
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def event_handler_ACTION_END(
        self, event: ActionEndEventArguments, match: Any
    ) -> List[DrawCardAction | RemoveObjectAction]:
        """
        If self action end and hand is empty, draw card.
        """
        if event.action.position.player_idx != self.position.player_idx:
            # not self player
            return []
        if len(match.player_tables[self.position.player_idx].hands) != 0:
            # not empty hand
            return []
        # draw a card
        self.usage -= 1
        return [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=1,
                draw_if_filtered_not_enough=True,
            )
        ] + self.check_should_remove()


class YayoiNanatsuki_4_1(MasterZhang_3_8):
    name: Literal["Yayoi Nanatsuki"]
    version: Literal["4.1"] = "4.1"
    card_cost_label: int = CostLabels.ARTIFACT.value


register_class(
    Paimon_3_3
    | Katheryne_3_6
    | Timaeus_3_3
    | Wagner_3_3
    | ChefMao_4_1
    | Tubby_3_3
    | Timmie_3_3
    | Liben_3_3
    | ChangTheNinth_3_3
    | Ellin_3_3
    | IronTongueTian_3_3
    | LiuSu_3_3
    | Hanachirusato_3_7
    | KidKujirai_3_7
    | Xudong_3_7
    | Dunyarzad_4_1
    | Rana_3_7
    | MasterZhang_3_8
    | Setaria_4_0
    | YayoiNanatsuki_4_1
    | Dunyarzad_3_7
    | Katheryne_3_3
    | ChefMao_3_3
)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...modifiable_values import CostValue, DamageValue
from ...event import (
    CreateObjectEventArguments,
    RemoveObjectEventArguments,
    SkillEndEventArguments,
    SwitchCharacterEventArguments,
)

from ...action import (
    ActionTypes,
    Actions,
    ChangeObjectUsageAction,
    MakeDamageAction,
    RemoveObjectAction,
)
from ...struct import Cost

from ...consts import (
    CostLabels,
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    SkillType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    CharacterBase,
    SkillTalent,
)


# Skills


class YuubanMeigen(ElementalNormalAttackBase):
    name: Literal["Yuuban Meigen"] = "Yuuban Meigen"
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = ElementalNormalAttackBase.get_cost(ElementType.ANEMO)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        if self has Windfavored, then deal 2 more damage and attack next
        character.
        """
        windfavored = None
        this_character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        for status in this_character.status:
            if status.name == "Windfavored":
                assert status.usage >= 1
                windfavored = status
                break
        if windfavored is None:
            # do normal attack
            return super().get_actions(match)
        self.damage = 3
        # increase damage, and attack next character
        ret = super().get_actions(match)
        self.damage = 1
        damage_action = ret[0]
        assert damage_action.type == ActionTypes.MAKE_DAMAGE
        enemy_table = match.player_tables[1 - self.position.player_idx]
        next = enemy_table.next_character_idx()
        if next is not None:
            # modify attack target
            damage_action.damage_value_list[0].target_position = enemy_table.characters[
                next
            ].position
        assert len(ret) == 2
        return [
            damage_action,
            ret[1],
            ChangeObjectUsageAction(
                object_position=windfavored.position, change_usage=-1
            ),
        ]


class HanegaSongOfTheWind(ElementalSkillBase):
    name: Literal["Hanega: Song of the Wind"] = "Hanega: Song of the Wind"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(elemental_dice_color=DieColor.ANEMO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(
            match,
            [
                self.create_character_status("Windfavored"),
            ],
        )


class KyougenFiveCeremonialPlays(ElementalBurstBase):
    name: Literal["Kyougen: Five Ceremonial Plays"] = "Kyougen: Five Ceremonial Plays"
    damage: int = 7
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ANEMO, elemental_dice_number=3, charge=3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        windfavored = None
        this_character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        for status in this_character.status:
            if status.name == "Windfavored":
                assert status.usage >= 1
                windfavored = status
                break
        if windfavored is not None:
            self.damage = 8
            ret = [
                RemoveObjectAction(object_position=windfavored.position)
            ] + super().get_actions(match)
            self.damage = 7
            return ret
        return super().get_actions(match)


# Talents


class GalesOfReverie_4_1(SkillTalent):
    name: Literal["Gales of Reverie"]
    version: Literal["4.1"] = "4.1"
    character_name: Literal["Wanderer"] = "Wanderer"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ANEMO,
        elemental_dice_number=4,
    )
    skill: Literal["Hanega: Song of the Wind"] = "Hanega: Song of the Wind"

    usage: int = 0
    switch_performed: bool = False

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[Actions]:
        """
        If self use normal attack, and has Windfavored, and is charged attack,
        mark talent activated.
        """
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
            source_area=ObjectPositionType.CHARACTER,
        ):
            # not self do skill, or not equipped
            return []
        if not (
            event.action.skill_type == SkillType.NORMAL_ATTACK
            and match.player_tables[self.position.player_idx].charge_satisfied
        ):
            # not charged attack
            return []
        this_character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        for status in this_character.status:
            if status.name == "Windfavored":
                # found windfavored, set card usage to 1
                self.usage = 1
        return []

    def value_modifier_COST(
        self,
        value: CostValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> CostValue:
        """
        If self has usage, reduce switch character cost from this character
        by 1
        """
        if self.usage <= 0:
            # not activated
            return value
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            source_area=ObjectPositionType.CHARACTER,
        ):
            # not this player, nor not equipped
            return value
        if value.cost.label & CostLabels.SWITCH_CHARACTER.value == 0:
            # not switch character
            return value
        # reduce 1 cost
        value.cost.decrease_cost(None)
        if mode == "REAL":
            # mark switch performed and decrease usage
            self.usage = 0
            self.switch_performed = True
        return value

    def event_handler_SWITCH_CHARACTER(
        self, event: SwitchCharacterEventArguments, match: Any
    ) -> List[Actions]:
        """
        If self has usage and switch from this character, make 1 anemo damage
        and decrease usage.
        """
        if not self.switch_performed:
            # not switch performed
            return []
        self.switch_performed = False
        target = match.player_tables[
            1 - self.position.player_idx
        ].get_active_character()
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        target_position=target.position,
                        damage=1,
                        damage_type=DamageType.DAMAGE,
                        damage_elemental_type=DamageElementalType.ANEMO,
                        cost=Cost(),
                    )
                ]
            )
        ]


# character base


class Wanderer_4_1(CharacterBase):
    name: Literal["Wanderer"]
    desc: Literal["", "transform"] = ""
    version: Literal["4.1"] = "4.1"
    element: ElementType = ElementType.ANEMO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[YuubanMeigen | HanegaSongOfTheWind | KyougenFiveCeremonialPlays] = []
    faction: List[FactionType] = []
    weapon_type: WeaponType = WeaponType.CATALYST

    def _init_skills(self) -> None:
        self.skills = [
            YuubanMeigen(),
            HanegaSongOfTheWind(),
            KyougenFiveCeremonialPlays(),
        ]

    def _update_desc(self, match):
        status = (
            match.player_tables[self.position.player_idx]
            .characters[self.position.character_idx]
            .status
        )
        for s in status:
            if s.name == "Windfavored":
                self.desc = "transform"
                return
        else:
            self.desc = ""

    def event_handler_REMOVE_OBJECT(
        self, event: RemoveObjectEventArguments, match: Any
    ) -> List[Actions]:
        self._update_desc(match)
        return []

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[Actions]:
        self._update_desc(match)
        return []


register_class(Wanderer_4_1 | GalesOfReverie_4_1)



from typing import Dict, List, Literal

from ....status.character_status.base import UsageCharacterStatus

from .....utils.class_registry import register_class
from ....character.character_base import SkillBase
from ....modifiable_values import CostValue, DamageIncreaseValue
from ....action import Actions, CreateDiceAction, CreateObjectAction, RemoveObjectAction
from ....match import Match
from ....event import (
    ReceiveDamageEventArguments,
    RoundEndEventArguments,
    RoundPrepareEventArguments,
    SkillEndEventArguments,
)
from ....consts import DamageType, IconType, ObjectPositionType, SkillType, WeaponType
from ....struct import Cost
from ....card.equipment.weapon.base import RoundEffectWeaponBase, WeaponBase
from .....utils.desc_registry import DescDictType


class LostPrayerToTheSacredWinds_4_3(WeaponBase):
    name: Literal["Lost Prayer to the Sacred Winds"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(same_dice_number=3)
    weapon_type: Literal[WeaponType.CATALYST] = WeaponType.CATALYST
    damage_increase: int = 0
    max_damage: int = 2

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Match
    ) -> List[Actions]:
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped
            return []
        self.damage_increase = min(self.damage_increase + 1, self.max_damage)
        return []


class TulaytullahsRemembrance_4_3(RoundEffectWeaponBase):
    name: Literal["Tulaytullah's Remembrance"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(same_dice_number=3)
    weapon_type: Literal[WeaponType.CATALYST] = WeaponType.CATALYST
    max_usage_per_round: int = 2

    def value_modifier_COST(
        self, value: CostValue, match: Match, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        If self use charged attack, and has usage, decrease one unaligned die
        cost.
        """
        if self.usage <= 0:
            # not enough usage, not modify
            return value
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
            source_area=ObjectPositionType.CHARACTER,
        ):
            # not this character use skill, or not equipped, not modify
            return value
        skill: SkillBase = match.get_object(value.position)  # type: ignore
        assert skill is not None
        skill_type: SkillType = skill.skill_type
        if skill_type != SkillType.NORMAL_ATTACK:
            # not normal attack, not modify
            return value
        if not match.player_tables[self.position.player_idx].charge_satisfied:
            # not charged attack, not modify
            return value
        # try decrease any cost
        value.cost.decrease_cost(None)
        if mode == "REAL":
            self.usage -= 1
        return value


class BeaconOfTheReedSeaStatus_4_3(UsageCharacterStatus):
    # TODO: in official implementation, two status are separated.
    name: Literal["Beacon of the Reed Sea"] = "Beacon of the Reed Sea"
    version: Literal["4.3"] = "4.3"
    usage: int = 1
    max_usage: int = 999

    icon_type: IconType = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Match, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        When equipped character is using skill, increase the damage.
        """
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            # not current character using skill
            return value
        # modify damage
        assert mode == "REAL"
        value.damage += self.usage
        self.usage = 0
        return value

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[RemoveObjectAction]:
        """
        In prepare, remove self
        """
        self.usage = 0
        return self.check_should_remove()


class BeaconOfTheReedSea_4_3(RoundEffectWeaponBase):
    name: Literal["Beacon of the Reed Sea"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(same_dice_number=3)
    weapon_type: Literal[WeaponType.CLAYMORE] = WeaponType.CLAYMORE
    max_usage_per_round: int = 1  # usage for elemental skill

    receive_damage_usage: int = 1
    max_receive_damage_usage_per_round: int = 1

    def equip(self, match: Match) -> List[Actions]:
        self.usage = self.max_usage_per_round
        self.receive_damage_usage = self.max_receive_damage_usage_per_round
        return []

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[Actions]:
        self.usage = self.max_usage_per_round
        self.receive_damage_usage = self.max_receive_damage_usage_per_round
        return []

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Match
    ) -> List[CreateObjectAction]:
        if self.receive_damage_usage == 0:
            # already used
            return []
        if not self.position.check_position_valid(
            event.final_damage.target_position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            area_same=True,
            source_area=ObjectPositionType.CHARACTER,
        ):
            # not equipped, or not self receive damage
            return []
        if event.final_damage.damage_type != DamageType.DAMAGE:
            # not damage
            return []
        # create status
        self.receive_damage_usage -= 1
        return [
            CreateObjectAction(
                object_name=self.name,
                object_position=self.position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                ),
                object_arguments={},
            )
        ]

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Match
    ) -> List[CreateObjectAction]:
        if self.usage == 0:
            # already used
            return []
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
            source_area=ObjectPositionType.CHARACTER,
        ):
            # not self, or not equipped
            return []
        skill: SkillBase = match.get_object(event.action.position)  # type: ignore
        assert skill is not None
        if skill.skill_type != SkillType.ELEMENTAL_SKILL:
            # not elemental skill
            return []
        self.usage -= 1
        return [
            CreateObjectAction(
                object_name=self.name,
                object_position=self.position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                ),
                object_arguments={},
            )
        ]


class PrimordialJadeWingedSpear_4_3(WeaponBase):
    name: Literal["Primordial Jade Winged-Spear"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(same_dice_number=3)
    weapon_type: Literal[WeaponType.POLEARM] = WeaponType.POLEARM
    max_damage: int = 3

    def equip(self, match: Match) -> List[Actions]:
        self.damage_increase = 1
        return []

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[Actions]:
        # reset damage increase
        self.damage_increase = 1
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Match
    ) -> List[CreateObjectAction]:
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
            source_area=ObjectPositionType.CHARACTER,
        ):
            # not self, or not equipped
            return []
        # increase damage
        self.damage_increase = min(self.damage_increase + 1, self.max_damage)
        return []


class LightOfFoliarIncision_4_3(RoundEffectWeaponBase):
    name: Literal["Light of Foliar Incision"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(same_dice_number=3)
    weapon_type: Literal[WeaponType.SWORD] = WeaponType.SWORD
    max_usage_per_round: int = 2

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Match
    ) -> List[CreateDiceAction]:
        if self.usage == 0:
            # out of usage
            return []
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
            source_area=ObjectPositionType.CHARACTER,
        ):
            # not self, or not equipped
            return []
        skill: SkillBase = match.get_object(event.action.position)  # type: ignore
        assert skill is not None
        if skill.skill_type != SkillType.NORMAL_ATTACK:
            # not normal attack
            return []
        self.usage -= 1
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx, number=1, different=True
            )
        ]


desc: Dict[str, DescDictType] = {
    "WEAPON/Lost Prayer to the Sacred Winds": {
        "names": {"en-US": "Lost Prayer to the Sacred Winds", "zh-CN": "四风原典"},
        "descs": {
            "4.3": {
                "en-US": 'For each point of "Bonus DMG" this card has, character deals +1 DMG.\nEnd Phase: This card gains 1 point of "Bonus DMG." (Max 2 points)\n(Can only be equipped by Catalyst Characters. Each Character can equip at most 1 Weapon.',  # noqa: E501
                "zh-CN": "此牌每有1点「伤害加成」，角色造成的伤害+1。\n结束阶段：此牌累积1点「伤害加成」。（最多累积到2点）\n（「法器」角色才能装备。角色最多装备1件「武器」）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Weapon_Yuandian.png",  # noqa: E501
        "id": 311106,
    },
    "WEAPON/Tulaytullah's Remembrance": {
        "names": {"en-US": "Tulaytullah's Remembrance", "zh-CN": "图莱杜拉的回忆"},
        "descs": {
            "4.3": {
                "en-US": "The character deals +1 DMG.\nWhen the character uses a Charged Attack: Spend 1 less Unaligned Element. (Can be triggered up to twice per Round)\n(Only Catalyst Characters can equip this. A character can equip a maximum of 1 Weapon)",  # noqa: E501
                "zh-CN": "角色造成的伤害+1。\n角色进行重击时：少花费1个无色元素。（每回合最多触发2次）\n（「法器」角色才能装备。角色最多装备1件「武器」）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Weapon_Tulaidu.png",  # noqa: E501
        "id": 311107,
    },
    "CHARACTER_STATUS/Beacon of the Reed Sea": {
        "names": {"en-US": "Beacon of the Reed Sea", "zh-CN": "苇海信标"},
        "descs": {
            "4.3": {
                "en-US": "Increase DMG dealt by character this Round the next time they deal DMG by the usage of this status.",  # noqa: E501
                "zh-CN": "本回合内，角色下次造成的伤害额外增加该状态的使用次数点。",  # noqa: E501
            }
        },
    },
    "WEAPON/Beacon of the Reed Sea": {
        "names": {"en-US": "Beacon of the Reed Sea", "zh-CN": "苇海信标"},
        "descs": {
            "4.3": {
                "en-US": "The character deals +1 DMG.\nAfter the character uses an Elemental Skill: +1 additional DMG dealt by character this Round the next time they deal DMG. (Once per Round)\nAfter this character takes DMG: +1 additional DMG dealt by character this Round the next time they deal DMG. (Once per Round)\n(Only Claymore Characters can equip this. A character can equip a maximum of 1 Weapon)",  # noqa: E501
                "zh-CN": "角色造成的伤害+1。\n角色使用「元素战技」后：本回合内，角色下次造成的伤害额外+1。（每回合1次）\n角色受到伤害后：本回合内，角色下次造成的伤害额外+1。（每回合1次）\n（「双手剑」角色才能装备。角色最多装备1件「武器」）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Weapon_Xinbiao.png",  # noqa: E501
        "id": 311306,
    },
    "WEAPON/Primordial Jade Winged-Spear": {
        "names": {"en-US": "Primordial Jade Winged-Spear", "zh-CN": "和璞鸢"},
        "descs": {
            "4.3": {
                "en-US": "The character deals +1 DMG.\nAfter this character uses a Skill: Until the end of this Round, +1 additional DMG provided by this card. (Up to +2 DMG total.)\n(Only Polearm Characters can equip this. A character can equip a maximum of 1 Weapon)",  # noqa: E501
                "zh-CN": "角色造成的伤害+1。\n角色使用技能后：直到回合结束前，此牌所提供的伤害加成值额外+1。（最多累积到+2）\n（「长柄武器」角色才能装备。角色最多装备1件「武器」）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Weapon_Hepuyuan.png",  # noqa: E501
        "id": 311407,
    },
    "WEAPON/Light of Foliar Incision": {
        "names": {"en-US": "Light of Foliar Incision", "zh-CN": "裁叶萃光"},
        "descs": {
            "4.3": {
                "en-US": "The character deals +1 DMG.\nAfter the character uses a Normal Attack: Create 1 random Elemental Die. (Can be triggered up to twice per Round)\n(Only Sword Characters can equip this. A character can equip a maximum of 1 Weapon)",  # noqa: E501
                "zh-CN": "角色造成的伤害+1。\n角色使用「普通攻击」后：生成1个随机类型的元素骰。（每回合最多触发2次）\n（「单手剑」角色才能装备。角色最多装备1件「武器」）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Weapon_Caiye.png",  # noqa: E501
        "id": 311506,
    },
}


register_class(
    LostPrayerToTheSacredWinds_4_3
    | TulaytullahsRemembrance_4_3
    | BeaconOfTheReedSea_4_3
    | PrimordialJadeWingedSpear_4_3
    | LightOfFoliarIncision_4_3
    | BeaconOfTheReedSeaStatus_4_3,
    desc,
)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import AttackerSummonBase

from ...modifiable_values import DamageValue
from ...event import MakeDamageEventArguments, RoundEndEventArguments

from ...action import (
    ActionTypes,
    Actions,
    ChangeObjectUsageAction,
    CreateDiceAction,
    MakeDamageAction,
)
from ...struct import Cost

from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE,
    ELEMENT_TO_DIE_COLOR,
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class GossamerSprite_4_2(AttackerSummonBase):
    name: Literal["Gossamer Sprite"] = "Gossamer Sprite"
    version: Literal["4.2"] = "4.2"
    usage: int = 1
    max_usage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.DENDRO
    damage: int = 1

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        ret = super().event_handler_ROUND_END(event, match)
        active_character = match.player_tables[
            self.position.player_idx
        ].get_active_character()
        make_damage_action = ret[0]
        assert make_damage_action.type == ActionTypes.MAKE_DAMAGE
        make_damage_action.damage_value_list.append(
            DamageValue(
                position=self.position,
                damage_type=DamageType.HEAL,
                target_position=active_character.position,
                damage=-1,
                damage_elemental_type=DamageElementalType.HEAL,
                cost=Cost(),
            )
        )
        return ret


# Skills


class UniversalDiagnosis(ElementalSkillBase):
    name: Literal["Universal Diagnosis"] = "Universal Diagnosis"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(elemental_dice_color=DieColor.DENDRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match, [self.create_summon("Gossamer Sprite")])


class HolisticRevivification(ElementalBurstBase):
    name: Literal["Holistic Revivification"] = "Holistic Revivification"
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(
        elemental_dice_color=DieColor.DENDRO, elemental_dice_number=4, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return [
            self.charge_self(-2),
            self.create_team_status("Pulsing Clarity"),
            self.create_team_status("Seamless Shield"),
        ]


# Talents


class AllThingsAreOfTheEarth_4_2(SkillTalent):
    name: Literal["All Things Are of the Earth"]
    version: Literal["4.2"] = "4.2"
    character_name: Literal["Baizhu"] = "Baizhu"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.DENDRO, elemental_dice_number=4, charge=2
    )
    skill: Literal["Holistic Revivification"] = "Holistic Revivification"

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[CreateDiceAction]:
        """
        When heal made by our shield, create one dice
        """
        source_position = event.action.damage_value_list[0].position
        if (
            self.position.area != ObjectPositionType.CHARACTER
            or source_position.area != ObjectPositionType.TEAM_STATUS
            or source_position.player_idx != self.position.player_idx
        ):
            # not our team status, or not equipped
            return []
        source = match.get_object(source_position)
        if source.name != "Seamless Shield":
            # not our shield
            return []
        # create dice
        active_character = match.player_tables[
            self.position.player_idx
        ].get_active_character()
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx,
                color=ELEMENT_TO_DIE_COLOR[active_character.element],
                number=1,
            )
        ]


# character base


class Baizhu_4_2(CharacterBase):
    name: Literal["Baizhu"]
    version: Literal["4.2"] = "4.2"
    element: ElementType = ElementType.DENDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase | UniversalDiagnosis | HolisticRevivification
    ] = []
    faction: List[FactionType] = [FactionType.LIYUE]
    weapon_type: WeaponType = WeaponType.CATALYST

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="The Classics of Acupuncture",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            UniversalDiagnosis(),
            HolisticRevivification(),
        ]


register_class(Baizhu_4_2 | AllThingsAreOfTheEarth_4_2 | GossamerSprite_4_2)



from typing import Any, List, Literal

from ....utils.class_registry import register_class
from ...modifiable_values import DamageValue

from ...summon.base import AttackerSummonBase

from ...event import ReceiveDamageEventArguments, SkillEndEventArguments

from ...action import Actions, ChangeObjectUsageAction, MakeDamageAction
from ...struct import Cost, ObjectPosition

from ...consts import (
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    ElementalReactionType,
    FactionType,
    ObjectPositionType,
    SkillType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class CryoCicins_4_1(AttackerSummonBase):
    name: Literal["Cryo Cicins"] = "Cryo Cicins"
    version: Literal["4.1"] = "4.1"
    usage: int = 2
    max_usage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.CRYO
    damage: int = 1
    renew_type: Literal["ADD"] = "ADD"
    decrease_only_self_damage: bool = True

    over_maximum: bool = False

    def renew(self, new_status: "CryoCicins_4_1") -> None:
        if self.usage + new_status.usage > self.max_usage:
            self.over_maximum = True
        return super().renew(new_status)

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[ChangeObjectUsageAction | MakeDamageAction]:
        if event.action.position.player_idx != self.position.player_idx:
            # not our player skill made damage, do nothing
            self.over_maximum = False
            return []
        assert event.action.position.area == ObjectPositionType.SKILL
        characters = match.player_tables[self.position.player_idx].characters
        character = characters[event.action.position.character_idx]
        if character.name != "Fatui Cryo Cicin Mage":
            # not Fatui Cryo Cicin Mage, do nothing
            self.over_maximum = False
            return []
        ret: List[ChangeObjectUsageAction | MakeDamageAction] = []
        if event.action.skill_type == SkillType.NORMAL_ATTACK:
            # change usage
            ret += [
                ChangeObjectUsageAction(
                    object_position=self.position,
                    change_usage=1,
                    max_usage=self.max_usage,
                )
            ]
            if self.usage + 1 > self.max_usage:
                self.over_maximum = True
        # if over_maximum and character has talent, 2 damage to opposite active
        if self.over_maximum and character.talent is not None:
            target_character = match.player_tables[
                1 - self.position.player_idx
            ].get_active_character()
            ret.append(
                MakeDamageAction(
                    damage_value_list=[
                        DamageValue(
                            position=self.position,
                            damage_type=DamageType.DAMAGE,
                            target_position=target_character.position,
                            damage=2,
                            damage_elemental_type=self.damage_elemental_type,
                            cost=Cost(),
                        )
                    ]
                )
            )
        self.over_maximum = False
        return ret

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Any
    ) -> List[ChangeObjectUsageAction]:
        if (
            event.final_damage.target_position.player_idx != self.position.player_idx
            or event.elemental_reaction == ElementalReactionType.NONE
            or event.final_damage.damage_type != DamageType.DAMAGE
        ):
            # not self player receive damage, or no elemental reaction,
            # or not damage
            return []
        target = match.player_tables[
            event.final_damage.target_position.player_idx
        ].characters[event.final_damage.target_position.character_idx]
        if self.decrease_only_self_damage and target.name != "Fatui Cryo Cicin Mage":
            # only decrease in self damage
            return []
        return [ChangeObjectUsageAction(object_position=self.position, change_usage=-1)]


# Skills


def find_cryo_cicins(position: ObjectPosition, match: Any, usage: int) -> Any | None:
    """
    Find cicins based on self position and match. then if with usage will
    exceed max usage, return cicin position. otherwise, return None.
    """
    summons = match.player_tables[position.player_idx].summons
    for summon in summons:
        if summon.name == "Cryo Cicins" and summon.usage + usage > summon.max_usage:
            return summon
    return None


class CicinIcicle(ElementalNormalAttackBase):
    name: Literal["Cicin Icicle"] = "Cicin Icicle"
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = ElementalNormalAttackBase.get_cost(ElementType.CRYO)


class MistySummons(ElementalSkillBase):
    name: Literal["Misty Summons"] = "Misty Summons"
    version: Literal["4.1"] = "4.1"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(
            match,
            [
                self.create_summon("Cryo Cicins", {"version": self.version}),
            ],
        )


class BlizzardBranchBlossom(ElementalBurstBase):
    name: Literal["Blizzard, Branch, Blossom"] = "Blizzard, Branch, Blossom"
    damage: int = 5
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.CRYO, elemental_dice_number=3, charge=3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        # max usage, so find regardless of its current usage
        cicin = find_cryo_cicins(self.position, match, 999)
        args = {}
        if cicin is not None:
            args["usage"] = 1 + cicin.usage
            args["max_usage"] = 1 + cicin.usage
        ret = [
            self.charge_self(-self.cost.charge),
            self.attack_opposite_active(
                match,
                self.damage,
                self.damage_type,
                [self.create_team_status("Flowing Cicin Shield", args)],
            ),
        ]
        app = self.element_application_self(match, DamageElementalType.CRYO)
        ret[1].damage_value_list += app.damage_value_list
        return ret


# Talents


class CicinsColdGlare_3_7(SkillTalent):
    name: Literal["Cicin's Cold Glare"]
    version: Literal["3.7"] = "3.7"
    character_name: Literal["Fatui Cryo Cicin Mage"] = "Fatui Cryo Cicin Mage"
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=3)
    skill: Literal["Misty Summons"] = "Misty Summons"


# character base


class FatuiCryoCicinMage_4_1(CharacterBase):
    name: Literal["Fatui Cryo Cicin Mage"]
    version: Literal["4.1"] = "4.1"
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[CicinIcicle | MistySummons | BlizzardBranchBlossom] = []
    faction: List[FactionType] = [FactionType.FATUI]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [CicinIcicle(), MistySummons(), BlizzardBranchBlossom()]


register_class(FatuiCryoCicinMage_4_1 | CryoCicins_4_1 | CicinsColdGlare_3_7)



from typing import Any, Dict, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import AttackerSummonBase

from ...modifiable_values import CostValue
from ...event import (
    ChooseCharacterEventArguments,
    GameStartEventArguments,
    RoundPrepareEventArguments,
    SwitchCharacterEventArguments,
)

from ...action import Actions, ChangeObjectUsageAction, CreateObjectAction
from ...struct import Cost, ObjectPosition

from ...consts import (
    CostLabels,
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    PassiveSkillBase,
    CharacterBase,
    TalentBase,
)


# Summons


class FrostflakeSekiNoTo_3_3(AttackerSummonBase):
    name: Literal["Frostflake Seki no To"] = "Frostflake Seki no To"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.CRYO
    damage: int = 2


# Skills


class KamisatoArtSoumetsu(ElementalBurstBase):
    name: Literal["Kamisato Art: Soumetsu"] = "Kamisato Art: Soumetsu"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.CRYO, elemental_dice_number=3, charge=3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match, [self.create_summon("Frostflake Seki no To")])


class KamisatoArtSenho(PassiveSkillBase):
    name: Literal["Kamisato Art: Senho"] = "Kamisato Art: Senho"

    def infuse(self, match: Any) -> List[CreateObjectAction]:
        args: Dict = {"mark": "Kamisato Ayaka"}
        if self.is_talent_equipped(match):
            args["talent_activated"] = True
        return [self.create_character_status("Cryo Elemental Infusion", args)]

    def event_handler_SWITCH_CHARACTER(
        self, event: SwitchCharacterEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        if (
            event.action.player_idx != self.position.player_idx
            or event.action.character_idx != self.position.character_idx
        ):
            # not switch to this character
            return []
        return self.infuse(match)

    def event_handler_CHOOSE_CHARACTER(
        self, event: ChooseCharacterEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        if (
            event.action.player_idx != self.position.player_idx
            or event.action.character_idx != self.position.character_idx
        ):
            # not choose this character
            return []
        return self.infuse(match)

    def event_handler_GAME_START(
        self, event: GameStartEventArguments, match: Any
    ) -> List[ChangeObjectUsageAction]:
        """
        As after choosing character in game start will be followed by a
        round preparing, to make infuse in game start, we increase its usage
        by 1 if it exists.
        """
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        for status in character.status:
            if (  # pragma: no branch
                status.name == "Cryo Elemental Infusion"
                and "mark" in status.__fields__
                and status.mark == "Kamisato Ayaka"
            ):
                # add one usage to avoid disappear
                return [
                    ChangeObjectUsageAction(
                        object_position=status.position, change_usage=1
                    )
                ]
        return []


# Talents


class KantenSenmyouBlessing_3_3(TalentBase):
    name: Literal["Kanten Senmyou Blessing"]
    version: Literal["3.3"] = "3.3"
    character_name: Literal["Kamisato Ayaka"] = "Kamisato Ayaka"
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=2)
    cost_label: int = (
        CostLabels.CARD.value | CostLabels.TALENT.value | CostLabels.EQUIPMENT.value
    )
    usage: int = 1
    max_usage: int = 1

    def is_valid(self, match: Any) -> bool:
        if self.position.area != ObjectPositionType.HAND:
            # not in hand, cannot equip
            raise AssertionError("Talent is not in hand")
        return len(self.get_targets(match)) > 0

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        All ayaka are targets
        """
        assert self.position.area != ObjectPositionType.CHARACTER
        ret: List[ObjectPosition] = []
        for c in match.player_tables[self.position.player_idx].characters:
            if c.name == self.character_name and c.is_alive:
                ret.append(c.position)
        return ret

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        Reset usage
        """
        self.usage = self.max_usage
        return []

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        If equipped, and has usage, and switch to ayaka, reduce 1 cost.
        """
        if self.usage <= 0:
            # out of usage, do nothing
            return value
        if value.cost.label & CostLabels.SWITCH_CHARACTER.value == 0:
            # not switch character, do nothing
            return value
        assert value.target_position is not None
        if not self.position.check_position_valid(
            value.target_position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            area_same=True,
            source_area=ObjectPositionType.CHARACTER,
        ):
            # not equipped or not switch to ayaka, do nothing
            return value
        if value.cost.decrease_cost(None):  # pragma: no branch
            # reduce 1 cost
            if mode == "REAL":
                self.usage -= 1
        return value


# character base


class KamisatoAyaka_3_3(CharacterBase):
    name: Literal["Kamisato Ayaka"]
    version: Literal["3.3"] = "3.3"
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase
        | ElementalSkillBase
        | KamisatoArtSoumetsu
        | KamisatoArtSenho
    ] = []
    faction: List[FactionType] = [FactionType.INAZUMA]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Kamisato Art: Kabuki",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            ElementalSkillBase(
                name="Kamisato Art: Hyouka",
                damage_type=DamageElementalType.CRYO,
                cost=ElementalSkillBase.get_cost(self.element),
            ),
            KamisatoArtSoumetsu(),
            KamisatoArtSenho(),
        ]


register_class(KamisatoAyaka_3_3 | FrostflakeSekiNoTo_3_3 | KantenSenmyouBlessing_3_3)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...event import MakeDamageEventArguments, SkillEndEventArguments

from ...summon.base import SwirlChangeSummonBase

from ...action import Actions, CreateObjectAction, SwitchCharacterAction
from ...struct import Cost, ObjectPosition

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    ElementalReactionType,
    FactionType,
    ObjectPositionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class AutumnWhirlwind_3_8(SwirlChangeSummonBase):
    name: Literal["Autumn Whirlwind"] = "Autumn Whirlwind"
    version: Literal["3.8"] = "3.8"
    usage: int = 3
    max_usage: int = 3
    damage: int = 1


# Skills


class Chihayaburu(ElementalSkillBase):
    name: Literal["Chihayaburu"] = "Chihayaburu"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(elemental_dice_color=DieColor.ANEMO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        ret: List[Actions] = []
        # make damage
        ret.append(
            self.attack_opposite_active(
                match,
                self.damage,
                self.damage_type,
                [self.create_character_status("Midare Ranzan: New")],
            )
        )
        # switch character
        next_character = match.player_tables[
            self.position.player_idx
        ].next_character_idx()
        if next_character is not None:
            ret.append(
                SwitchCharacterAction(
                    player_idx=self.position.player_idx,
                    character_idx=next_character,
                )
            )
        ret.append(self.charge_self(1))
        return ret


class KazuhaSlash(ElementalBurstBase):
    name: Literal["Kazuha Slash"] = "Kazuha Slash"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ANEMO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(
            match,
            [
                self.create_summon("Autumn Whirlwind"),
            ],
        )


# Talents


class PoeticsOfFuubutsu_3_8(SkillTalent):
    name: Literal["Poetics of Fuubutsu"]
    version: Literal["3.8"] = "3.8"
    character_name: Literal["Kaedehara Kazuha"] = "Kaedehara Kazuha"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ANEMO,
        elemental_dice_number=3,
    )
    skill: Literal["Chihayaburu"] = "Chihayaburu"

    swirl_element = ElementType.NONE

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        when damage made, if it is from kazuha, and has swirl, record element.
        """
        for damage in event.damages:
            if not self.position.check_position_valid(
                damage.final_damage.position,
                match,
                player_idx_same=True,
                character_idx_same=True,
                target_area=ObjectPositionType.SKILL,
            ):
                # not self character use skill, do nothing
                continue
            if damage.final_damage.damage_from_element_reaction:
                # damage from elemental reaction, do nothing
                continue
            if damage.elemental_reaction != ElementalReactionType.SWIRL:
                # not swirl, do nothing
                continue
            assert (
                self.swirl_element == ElementType.NONE
            ), "element type already determined"
            # swirl, change element type
            elements = damage.reacted_elements
            assert elements[0] == ElementType.ANEMO, "First must be anemo"
            self.swirl_element = elements[1]
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[Actions]:
        """
        If kazuha did skill and swirled, generate Poetics of Fuubutsu with
        corresponding element.
        """
        element = self.swirl_element
        # regardless of skill situation, reset swirl element
        self.swirl_element = ElementType.NONE
        if element == ElementType.NONE:
            # no swirl, do nothing
            return []
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
            source_area=ObjectPositionType.CHARACTER,
        ):  # pragma: no cover
            # not same player or character, or not skill, or not equipped
            return []
        # this character swirl, generate Poetics of Fuubutsu
        return [
            CreateObjectAction(
                object_name=f"{self.name}: {element.name.capitalize()}",
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=0,
                ),
                object_arguments={},
            )
        ]


# character base


class KaedeharaKazuha_3_8(CharacterBase):
    name: Literal["Kaedehara Kazuha"]
    version: Literal["3.8"] = "3.8"
    element: ElementType = ElementType.ANEMO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[PhysicalNormalAttackBase | Chihayaburu | KazuhaSlash] = []
    faction: List[FactionType] = [FactionType.INAZUMA]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Garyuu Bladework",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            Chihayaburu(),
            KazuhaSlash(),
        ]


register_class(KaedeharaKazuha_3_8 | PoeticsOfFuubutsu_3_8 | AutumnWhirlwind_3_8)
